from __future__ import annotations
import io
import json
import os
import sys
import traceback
from typing import List, Optional, Tuple
import yaml
from hms_utils.argv import ARGV, AT_LEAST_ONE_OF, AT_MOST_ONE_OF, DEPENDENCY, DEPENDS_ON, OPTIONAL, REQUIRED   # noqa
from hms_utils.chars import chars
from hms_utils.config.config import Config
from hms_utils.config.config_output import ConfigOutput
from hms_utils.config.config_with_aws_macros import ConfigWithAwsMacros
from hms_utils.crypt_utils import read_encrypted_file
from hms_utils.dictionary_parented import JSON
from hms_utils.path_utils import is_current_or_parent_relative_path
from hms_utils.type_utils import any_of_bool, at_most_one_of_bool
from hms_utils.version_utils import get_version

DEFAULT_CONFIG_DIR = "~/.config/hms"
DEFAULT_CONFIG_FILE_NAME = "config.json"
DEFAULT_SECRETS_FILE_NAME = "secrets.json"
DEFAULT_PATH_SEPARATOR = "/"
OBFUSCATED_VALUE = "********"


def main(argv: Optional[List] = None):

    args = parse_args(argv if isinstance(argv, list) else sys.argv[1:])

    config = args.config
    merged_paths, unmerged_paths = config.merge(args.configs_for_merge)
    config.include(args.configs_for_include)

    if args.noaws:
        config._noaws = True

    if not args.lookup_paths and (args.tree or args.list or args.dump):
        if config.name:
            print(f"{chars.rarrow} {config.name}"
                  f"{f' {chars.dot} decrypted' if config.decrypted else ''}")
        if args.configs_for_merge:
            for config_for_merge in args.configs_for_merge:
                print(f"{chars.rarrow_hollow} {config_for_merge.name} {chars.dot} merged"
                      f"{f' {chars.dot} decrypted' if config_for_merge.decrypted else ''}")
        if args.configs_for_include:
            for config_for_include in args.configs_for_include:
                print(f"{chars.rarrow_hollow} {config_for_include.name} {chars.dot} included"
                      f"{f' {chars.dot} decrypted' if config_for_include.decrypted else ''}")

    if not args.lookup_paths:
        if args.json:
            if args.show is True:
                data = config.lookup("/", show=args.show).sorted()
            elif args.show is False:
                data = config.data(show=False)
            else:
                data = config.data(show=None)
            print(json.dumps(data, indent=4 if args.formatted else None))
        elif args.tree or args.dump:
            ConfigOutput.print_tree(config, show=args.show, nocolor=args.nocolor, root=True, debug=args.dump)
        elif args.list:
            ConfigOutput.print_list(config, show=args.show, nocolor=args.nocolor)

    status = 0
    if args.lookup_paths:
        if args.exports:
            status = handle_exports_command(config, args)
        else:
            status = handle_lookup_command(config, args)

    if config._warnings:
        if args.warnings:
            print(f"{chars.rarrow} WARNINGS ({len(config._warnings)}):", file=sys.stderr)
            for warning in config._warnings:
                print(f"  {chars.rarrow_hollow} {warning}", file=sys.stderr)

    if args.new:
        parse_args_new()

    sys.exit(status)


def main_show_script_path():
    sys.argv = ["hmsconfig", "--functions"]
    main()


def handle_lookup_command(config: Config, args: object) -> int:
    if not args.lookup_paths:
        return 0
    status = 0 ; n = 0  # noqa
    for lookup_path in args.lookup_paths:
        if (value := config.lookup(lookup_path, show=args.show if not args.dump else None)) is None:
            status = 1
            if not args.verbose:
                continue
            value = chars.null
        if args.show and Config._contains_macro(value):
            status = 1
        if isinstance(value, JSON):
            if args.json and args.formatted:
                value = json.dumps(value.sorted(), indent=4)
            elif args.tree or args.dump:
                prefix = "...\n" if args.verbose else ("" if n == 0 else "\n")
                root = value.context_path(path_separator=config.path_separator, path_rooted=True) if args.dump else None
                value = (prefix +
                         ConfigOutput.print_tree(config, data=value.sorted(), nocolor=args.nocolor,
                                                 string=True, indent=2 if args.verbose else None,
                                                 show=args.show, root=root, debug=args.dump))
        if args.verbose:
            print(f"{lookup_path}: {value}")
        else:
            print(f"{value}")
        n += 1
    return status


def handle_exports_command(config: Config, args: object) -> int:
    exports, status = config.exports(args.lookup_paths, show=args.show)
    for export_key in exports:
        if Config._contains_macro(exports[export_key]):
            status = 1
            break
    if args.exports_file:
        if os.path.exists(args.exports_file):
            _error(f"Export file must not already exist: {args.exports_file}")
        with io.open(args.exports_file, "w") as f:
            for export in exports:
                export = f"export {export}={exports[export]}"
                f.write(f"{export}\n")
                if args.verbose:
                    print(f"{chars.rarrow_hollow} {export}")
    else:
        if args.json:
            print(json.dumps(exports, indent=4 if args.formatted else None))
        else:
            for export in sorted(exports):
                export = f"export {export}={exports[export]}"
                print(export)
    return status


def parse_args_new() -> object:
    argv = ARGV({
        OPTIONAL([str]): ["--configs", "--config", "--confs", "--conf",
                          "--merge", "--merge-configs", "--merge-config", "--merge-confs", "--merge-conf"],
        OPTIONAL([str]): ["--secrets", "--secret", "--merge-secrets", "--merge-secret"],
        OPTIONAL([str]): ["--includes", "--include", "--include-configs", "--include-config",
                          "--include-confs", "--include-conf", "--imports", "--import",
                          "--import-configs", "--import-config", "import-confs", "--import-conf"],
        OPTIONAL(str): ["--config-dir", "--conf-dir", "--dir", "--directory"],
        OPTIONAL([str]): ["--lookup-paths", "--lookup-path", "--lookups", "--lookup"],
        OPTIONAL(bool): ["--tree"],
        OPTIONAL(bool): ["--list"],
        OPTIONAL(bool): ["--dump"],
        OPTIONAL(bool): ["--raw"],
        OPTIONAL(bool): ["--show"],
        OPTIONAL(str): ["--identity"],
        OPTIONAL(str): ["--password"],
        OPTIONAL(bool): ["--json"],
        OPTIONAL(bool): ["--format", "--formatted"],
        OPTIONAL(bool): ["--exports", "--export"],
        OPTIONAL(bool): ["--functions", "--function", "--shell"],
        OPTIONAL(str): ["--exports-file", "--export-file"],
        OPTIONAL(bool): ["--noaws"],
        OPTIONAL(str): ["--aws-profile", "--aws", "--profile", "--env"],
        OPTIONAL(bool): ["--nocolor"],
        OPTIONAL(bool): ["--warnings", "--warning"],
        OPTIONAL(bool): ["--verbose"],
        OPTIONAL(bool): ["--debug"],
        OPTIONAL(bool): ["--version"],
        OPTIONAL(bool): ["--new"],
        AT_MOST_ONE_OF: ["--tree", "--list", "--json", "--dump", "--exports", "--functions"],
        DEPENDENCY: ["--format", DEPENDS_ON, "--json"]
    }, parse=True, exit=False)

    argv.parse(exit=False)
    print(">>> NEW ARGV PARSING:")
    print(json.dumps(argv._dict, indent=4))


def parse_args(argv: List[str]) -> object:

    # TODO: Use to ARGV class ...

    class Args:
        config_dir = None
        config = None
        configs_for_merge = []
        configs_for_include = []
        lookup_paths = []
        json = False
        exports = False
        exports_file = None
        tree = False
        list = False
        files = False
        dump = False
        raw = False
        check = False
        show = None
        profile = None
        noaws = False
        nocolor = False
        formatted = False
        verbose = False
        password = False
        warnings = False
        debug = False
        new = False

    args = Args()

    def get_configs(_merges: bool = False, _includes: bool = False) -> None:

        nonlocal argv, args

        # We allow one or more files to be listed after --config or --secrets. Only the first file
        # in the list may NOT end in a ".json" suffix (as a argument not end in that, or and argument
        # starting with a dash, signifies the and of this list). If a file name contains "secret",
        # or is prepended with a special "secret:" or "secrets:" prefix, or if it is listed afer
        # the --secrets option, then it is assumed to contain secrets, which is really just ensures
        # that the values therein won't displayed by default when listing the contents of the file.
        # Multiple --config or --secrets options can be used; so you have lots of options option-wise.

        # If and only if either of the the --config or --secrets options are given then the
        # default config/secrets file (e.g. i.e. in  the ~/.config/hms directory) will NOT be used.

        def read_file(file: str) -> Tuple[Optional[dict], bool]:
            if args.password:
                try:
                    if data := read_encrypted_file(file, password=args.password):
                        if file.endswith(".yaml") or config_file.endswith(".yml"):
                            return yaml.safe_load(data), True
                        else:
                            return json.loads(data), True
                except Exception:
                    pass
            with io.open(file, "r") as f:
                if file.endswith(".yaml") or file.endswith(".yml"):
                    data = yaml.safe_load(f)
                else:
                    data = json.load(f)
            return data, False

        def get_config_dir() -> None:
            nonlocal argv, args
            get_password_arg()
            config_dir = os.path.expanduser(DEFAULT_CONFIG_DIR)
            if value := os.environ.get("HMS_CONFIG_DIR"):
                config_dir = value
            argi = 0 ; argn = len(argv)  # noqa
            while argi < argn:
                arg = argv[argi] ; argi += 1  # noqa
                if arg in ["--dir", "-dir", "--directory", "-directory"]:
                    if not ((argi < argn) and (config_dir := argv[argi])):
                        _usage()
                    del argv[argi - 1:argi + 1]
                    break
            if not config_dir:
                return None
            if not os.path.isabs(config_dir):
                config_dir = os.path.normpath(os.path.join(os.getcwd(), config_dir))
            if not os.path.isdir(config_dir := os.path.normpath(config_dir)):
                _error(f"Configuration directory does not exist: {config_dir}")
            args.config_dir = config_dir

        def verify_config(config_file: str, config_dir: str, secrets: bool = False) -> Config:  # noqa
            if not secrets:
                if config_file.startswith("secret:"):
                    secrets = True
                    config_file = config_file[len("secret:"):]
                elif config_file.startswith("secrets:"):
                    secrets = True
                    config_file = config_file[len("secrets:"):]
                elif "secret" in config_file.lower():
                    secrets = True
            config_file = os.path.expanduser(config_file)
            if os.path.isabs(config_file):
                config_file = os.path.normpath(config_file)
            elif is_current_or_parent_relative_path(config_file) and os.path.isfile(config_file):
                config_file = os.path.normpath(os.path.join(os.getcwd(), config_file))
            elif os.path.isfile(file := os.path.normpath(os.path.join(config_dir, config_file))):
                config_file = file
            else:
                config_file = os.path.normpath(os.path.join(os.getcwd(), config_file))
            if not os.path.isfile(config_file):
                _error(f"Configuration file does not exist: {config_file}")
            try:
                config_json, decrypted = read_file(config_file)
                config = Config(config_json, name=config_file, secrets=secrets, decrypted=decrypted)
            except Exception:
                _error(f"Configuration JSON file cannot be loaded: {config_file}")
            return config

        if not args.config_dir:
            get_config_dir()
        config_dir = args.config_dir

        configs = []
        if not (config_dir_option_specified := isinstance(config_dir, str) and config_dir):
            config_dir = os.path.expanduser(DEFAULT_CONFIG_DIR)
        argi = 0 ; argn = len(argv)  # noqa
        while argi < argn:
            arg = argv[argi] ; argi += 1  # noqa
            arg_config = arg in ["--config", "-config", "--conf", "-conf"]
            arg_secrets = arg in ["--secrets", "-secrets", "--secret", "-secret"]
            arg_merge_config = _merges and arg in ["--merge", "-merge"]
            arg_merge_secrets = _merges and arg in ["--merge-secrets", "-merge-secrets",
                                                    "--merge-secret", "-merge-secret"]
            arg_include_config = _includes and arg in ["--includes", "-includes", "--include", "-include",
                                                       "--imports", "-imports", "--import", "-import",
                                                       "--import-config", "-import-config",
                                                       "--import-configs", "-import-configs"]
            arg_include_secrets = _includes and arg in ["--include-secrets", "-include-secrets",
                                                        "--include-secret", "-include-secret",
                                                        "--import-secrets", "-import-secrets",
                                                        "--import-secret", "-import-secret"]
            if (arg_config or arg_secrets or
                arg_merge_config or arg_merge_secrets or arg_include_config or arg_include_secrets):  # noqa
                if not config_dir_option_specified:
                    config_dir = os.getcwd()
                secrets = arg_secrets or arg_include_secrets
                argi_config = argi - 1
                if not ((argi < argn) and (config_file := argv[argi])):
                    _usage()
                if not is_current_or_parent_relative_path(config_file):
                    config_file = os.path.join(".", config_file)
                configs.append(verify_config(config_file, config_dir, secrets=secrets))
                argi += 1
                while argi < argn:
                    arg = argv[argi]
                    if (arg.startswith("-") or (not ((config_file := arg).endswith(".json") or
                                                     config_file.endswith(".yaml") or config_file.endswith(".yml")))):  # noqa
                        del argv[argi_config:argi] ; argi = 0 ; argn = len(argv)  # noqa
                        break
                    if not is_current_or_parent_relative_path(config_file):
                        config_file = os.path.join(".", config_file)
                    configs.append(verify_config(config_file, config_dir, secrets=secrets))
                    argi += 1
                if argi > 0:
                    del argv[argi_config:argi + 1]
        if _merges:
            args.configs_for_merge += configs
        elif _includes:
            args.configs_for_include = configs
        else:
            if not configs:
                configs.append(verify_config(DEFAULT_CONFIG_FILE_NAME, config_dir, secrets=False))
                configs.append(verify_config(DEFAULT_SECRETS_FILE_NAME, config_dir, secrets=True))
            args.config = configs[0]
            args.configs_for_merge = configs[1:]
            get_configs(_merges=True)
            get_configs(_includes=True)

    def get_lookup_paths() -> List[str]:
        nonlocal argv, args
        lookup_paths = []
        argi_lookup_paths = None
        argi_end_lookup_paths = len(argv)
        argi = 0 ; argn = len(argv)  # noqa
        while argi < argn:
            arg = argv[argi] ; argi += 1  # noqa
            if arg in ["--lookup", "-lookup"]:
                argi_lookup_paths = argi - 1
                continue
            elif arg.startswith("-") and argi_lookup_paths is not None:
                argi_end_lookup_paths = argi - 1
                break
            elif argi_lookup_paths is not None:
                lookup_paths.append(arg)
        if argi_lookup_paths is not None:
            del argv[argi_lookup_paths:argi_end_lookup_paths]
        args.lookup_paths = lookup_paths

    def get_password_arg():
        nonlocal argv, args
        argi = 0 ; argn = len(argv)  # noqa
        while argi < argn:
            arg = argv[argi].strip() ; argi += 1  # noqa
            if arg in ["--password", "-password", "--passwd", "-passwd"]:
                if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                    _usage()
                args.password = argv[argi]
                del argv[argi - 1:argi + 1]
                break

    def get_other_args():
        nonlocal argv, args
        argi = 0 ; argn = len(argv)  # noqa
        while argi < argn:
            arg = argv[argi].strip() ; argi += 1  # noqa
            if arg in ["--shell", "-shell", "--script", "-script", "--scripts", "-scripts",
                       "--command", "-command", "--commands", "-commands",
                       "--function", "-function", "--functions", "-functions"]:
                print(os.path.join(os.path.dirname(os.path.abspath(__file__)), "config_cli.sh"))
                exit(0)
            elif arg in ["--show", "-show"]:
                args.show = True
            elif arg in ["--identity", "-identity", "--aws-secrets-name", "-aws-secrets-name"]:
                if not ((argi < argn) and (arg := argv[argi].strip())):
                    _usage()
                os.environ[ConfigWithAwsMacros._AWS_SECRET_NAME_NAME] = arg ; argi += 1  # noqa
            elif arg in ["--list", "-list"]:
                args.list = True
            elif arg in ["--tree", "-tree"]:
                args.tree = True
            elif arg in ["--dump", "-dump"]:
                args.dump = True
            elif arg in ["--raw", "-raw"]:
                args.raw = True
            elif arg in ["--verbose", "-verbose"]:
                args.verbose = True
            elif arg in ["--debug", "-debug"]:
                args.debug = True
            elif arg in ["--check", "-check"]:
                args.check = True
            elif arg in ["--nocolor", "-nocolor"]:
                args.nocolor = True
            elif arg in ["--noaws", "-noaws"]:
                args.noaws = True
            elif arg in ["--warnings", "-warnings", "--warning", "-warning"]:
                args.warnings = True
            elif arg in ["--format", "-format", "--formatted", "-formatted"]:
                args.formatted = True
            elif arg in ["--new"]:
                args.new = True
            elif arg in ["--password", "-password", "--passwd", "-passwd"]:
                if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                    _usage()
                pass
                args.password = argv[argi] ; argi += 1  # noqa
            elif arg in ["--version", "-version"]:
                print(get_version())
                exit(0)
            elif arg in ["--aws", "-aws", "--aws", "-aws",
                         "--aws-profile", "-aws-profile", "--profile", "-profile", "--env", "-env"]:
                if (argi >= argn) or not (arg := argv[argi].strip()) or (not arg):
                    _usage()
                args.profile = arg ; argi += 1  # noqa
            elif arg in ["--json", "-json"]:
                args.json = True
            elif arg in ["--jsonf", "-jsonf"]:
                args.json = True
                args.formatted = True
            elif arg in ["--export", "-export", "--exports", "-exports"]:
                args.exports = True
                args.show = True
            elif arg in ["--export-file", "-export-file", "--exports-file", "-exports-file"]:
                if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                    _usage()
                args.exports_file = argv[argi] ; argi += 1  # noqa
                args.exports = True
                args.show = True
            elif arg.startswith("-"):
                _usage()
            else:
                args.lookup_paths.append(arg)

    get_configs()
    get_lookup_paths()
    get_other_args()

    if args.lookup_paths:
        if any_of_bool(args.list):
            _usage()
    else:
        if args.exports:
            _usage()
        if not at_most_one_of_bool(args.json, args.tree, args.list, args.dump):
            _usage()
    if not at_most_one_of_bool(args.tree, args.list, args.json, args.dump):
        _usage()
    if args.formatted:
        if args.tree or args.list or args.dump:
            _usage()
    if args.exports:
        if args.tree or args.dump:
            _usage()
    if not (args.lookup_paths or args.tree or args.json or args.list or args.dump):
        args.tree = True
    if args.raw:
        if args.show is not None:
            _usage()
        args.show = None
    elif args.show is None:
        args.show = False
    if args.show is False:
        args.noaws = True

    if args.profile:
        os.environ[ConfigWithAwsMacros._AWS_PROFILE_ENV_NAME] = args.profile
        if not args.config._aws_current_account_number(args.profile):
            _error(f"Specified AWS profile does not work: {args.profile}")

    return args


def _error(message: str, usage: bool = False, status: int = 1,
           exception: Optional[Exception] = None, trace: bool = False) -> None:
    print(f"ERROR: {message}", file=sys.stderr)
    if usage:
        _usage()
    if isinstance(exception, Exception):
        print(str(exception))
    if trace:
        traceback.print_exc()
    sys.exit(status)


def _usage():
    print("USAGE: hmsconfig OPTIONS [path]")
    print("OPTIONS:")
    print("--config:  list of JSON config files")
    print("--tree:    show all config data in tree format (default)")
    print("--json:    show all config data in json format")
    print("--list:    show all config data in list format")
    print("--dump:    show all config data in demp/debug format")
    print("--verbose: verbose output")
    print("--debug:   debugging output")
    sys.exit(1)


if __name__ == "__main__":
    main()



from copy import deepcopy
from functools import lru_cache
import json
import os
import sys
from typing import List, Optional, Union
import unicodedata
from dcicutils.structured_data import StructuredDataSet
from dcicutils.portal_utils import Portal
from hms_utils.dictionary_utils import sort_dictionary
from hms_utils.chars import chars


class PROPERTY:
    LAST_NAME = "last_name"
    FIRST_NAME = "first_name"
    EMAIL = "email"
    SUBMISSION_CENTER = "submission_center"
    SUBMISSION_CENTERS = "submission_centers"
    SUBMITS_FOR = "submits_for"
    REVOKED = "revoked"
    STATUS = "status"


USER_SPREADSHEET_COLUMN_MAP = {
    "Dana Farber Cancer Institute, Harvard Medical School": None,
    "SMaHT Listed Last Name": PROPERTY.LAST_NAME,
    "SMaHT Listed First Name": PROPERTY.FIRST_NAME,
    "Email": PROPERTY.EMAIL,
    "SMaHT Contact PI Association": None,
    "Grant Component": None,
    "DAC code in the portal": PROPERTY.SUBMISSION_CENTER,
    "Data submitter": PROPERTY.SUBMITS_FOR,
    "Revoked": PROPERTY.REVOKED
}


def sanity_check_users_from_spreadsheet_with_portal(file_or_users: Union[str, List[dict]],
                                                    portal_or_portal_env: Union[str, Portal],
                                                    verbose: bool = False,
                                                    debug: bool = False) -> None:

    def print_diffs_by_kind(diffs_by_kind: dict) -> None:
        if diffs_by_kind:
            for diff in diffs_by_kind:
                print(f"{chars.rarrow_hollow} {diff}")
                for user in diffs_by_kind[diff]:
                    print(f"  - {user['email']} {chars.dot} {user['uuid']}")

    diffs_by_kind = {}

    if isinstance(portal_or_portal_env, str):
        portal = Portal(portal_or_portal_env)
    elif not isinstance(portal := portal_or_portal_env, Portal):
        return
    if isinstance(file_or_users, str):
        users = read_users_from_spreadsheet(file_or_users)
    elif not isinstance(users := file_or_users, list):
        return

    users_to_add = []
    users_to_update = []
    for user in users:
        try:
            user_metadata = portal_get_user(portal, user[PROPERTY.EMAIL])
            if not (user_portal := get_user_from_portal_metadata(user_metadata)):
                _info(f"User from spreadsheet NOT FOUND in portal: {user[PROPERTY.EMAIL]}")
                if verbose:
                    print(json.dumps(user, indent=4))
                continue
            if debug is True:
                _debug(f"Read user from portal OK: {user[PROPERTY.EMAIL]}")
            if users_are_equal(user, user_portal):
                if verbose is True:
                    _info(f"User from spreadsheet and portal are the SAME: {user[PROPERTY.EMAIL]}")
            else:
                users_to_update.append(user)
                _error(f"User from spreadsheet and portal are DIFFERENT:"
                       f" {user[PROPERTY.EMAIL]} {chars.dot} {user_portal['uuid']}")
                diffs = compile_diffs(user, user_portal)
                for diff in diffs:
                    if diff not in diffs_by_kind:
                        diffs_by_kind[diff] = []
                    diffs_by_kind[diff].append({"email": user[PROPERTY.EMAIL], "uuid": user_portal["uuid"]})
                if debug is True:
                    _error(f"{chars.rarrow} SPREADSHEET: {user}")
                    _error(f"{chars.rarrow} PORTAL:      {user_portal}")
        except Exception as e:
            if ("not" in str(e).lower()) and ("found" in str(e).lower()):
                _error(f"User from spreadsheet is MISSING in portal: {user[PROPERTY.EMAIL]}")
                users_to_add.append(user)
            else:
                _error(f"Cannot read user from portal: {user[PROPERTY.EMAIL]}")
                _error(str(e))

    print_diffs_by_kind(diffs_by_kind)
    return users_to_add, users_to_update


def read_users_from_spreadsheet(file: str):
    data = StructuredDataSet(file)
    data = data.data
    data = next(iter(data.values()))
    users = []
    for item in data:
        users_from_spreadsheet_row = get_user_from_spreadsheet_row(item)
        for user_from_spreadsheet_row in users_from_spreadsheet_row:
            users.append(user_from_spreadsheet_row)
    return users


def get_user_from_spreadsheet_row(row: object) -> List[dict]:
    # Returns zero or more user dictionaries for the given spreadsheet row.
    # Can return more than one if there are more than one email in the email field.
    user = {}
    for spreadsheet_column_name in USER_SPREADSHEET_COLUMN_MAP:
        if user_column_name := USER_SPREADSHEET_COLUMN_MAP.get(spreadsheet_column_name):
            if spreadsheet_column_value := row.get(spreadsheet_column_name):
                user[user_column_name] = normalize_string(spreadsheet_column_value)
            else:
                user[user_column_name] = ""
    user[PROPERTY.EMAIL] = user[PROPERTY.EMAIL].lower()
    user[PROPERTY.SUBMISSION_CENTERS] = map_spreadsheet_submission_centers(user[PROPERTY.SUBMISSION_CENTER])
    del user[PROPERTY.SUBMISSION_CENTER]
    if user[PROPERTY.SUBMITS_FOR].lower() in ["yes", "y", "true"]:
        user[PROPERTY.SUBMITS_FOR] = user[PROPERTY.SUBMISSION_CENTERS]
    else:
        user[PROPERTY.SUBMITS_FOR] = []
    user[PROPERTY.STATUS] = "deleted" if (user[PROPERTY.REVOKED].lower() in ["yes", "y", "true"]) else "current"
    del user[PROPERTY.REVOKED]
    users = [user := sort_dictionary(user)]
    if "," in user[PROPERTY.EMAIL]:
        # Handle multiple (comma-separated) emails within a single field;
        # e.g.: abyzov.alexej@mayo.edu, alexej.abyzov@yale.edu
        if len(user_emails := list(set([item.strip() for item in user["email"].split(",")]))) > 1:
            user[PROPERTY.EMAIL] = user_emails[0]
            for user_email in user_emails[1:]:
                user = deepcopy(user)
                user[PROPERTY.EMAIL] = user_email.lower()
                users.append(user)
    if not user[PROPERTY.SUBMISSION_CENTERS]:
        del user[PROPERTY.SUBMISSION_CENTERS]
    if not user[PROPERTY.SUBMITS_FOR]:
        del user[PROPERTY.SUBMITS_FOR]
    return users


def map_spreadsheet_submission_centers(submission_centers: List[str]) -> List[str]:
    mapped_submission_centers = []
    if isinstance(submission_centers, str):
        submission_centers = submission_centers.split(",")
    if isinstance(submission_centers, list):
        for submission_center in submission_centers:
            if isinstance(submission_center, str) and (submission_center := submission_center.strip()):
                if submission_center.lower() == "dac":
                    submission_center = "smaht_dac"
                elif submission_center.lower() == "nih":
                    submission_center = None
                if submission_center and (submission_center not in mapped_submission_centers):
                    mapped_submission_centers.append(submission_center)
    return sorted(mapped_submission_centers)


def get_user_from_portal_metadata(user_metadata: dict) -> Optional[dict]:
    if isinstance(user_metadata, dict):
        user = {}
        user[PROPERTY.EMAIL] = normalize_string(user_metadata.get("email", ""))
        user[PROPERTY.FIRST_NAME] = normalize_string(user_metadata.get("first_name", ""))
        user[PROPERTY.LAST_NAME] = normalize_string(user_metadata.get("last_name", ""))
        # user[PROPERTY.REVOKED] = user_metadata.get("status", "").lower() in ["deleted", "revoked"]
        user[PROPERTY.STATUS] = user_metadata.get("status", "")
        user["uuid"] = user_metadata.get("uuid", "")
        if isinstance(submission_centers := user_metadata.get("submission_centers", []), list):
            user[PROPERTY.SUBMISSION_CENTERS] = \
                sorted(list(set([item.get("identifier", "") for item in submission_centers if item.get("identifier")])))
        if isinstance(submits_for := user_metadata.get("submits_for", []), list):
            user[PROPERTY.SUBMITS_FOR] = \
                sorted(list(set([item["identifier"] for item in submits_for if item.get("identifier")])))
        if not user[PROPERTY.SUBMISSION_CENTERS]:
            del user[PROPERTY.SUBMISSION_CENTERS]
        if not user[PROPERTY.SUBMITS_FOR]:
            del user[PROPERTY.SUBMITS_FOR]
        return sort_dictionary(user)
    return None


def get_nth_value_in_dictionary(data: dict, nth: int) -> str:
    if isinstance(data, dict) and isinstance(nth, int) and nth >= 0:
        try:
            if nth < len(keys := list(data.keys())):
                if isinstance(value := data[keys[nth]], str):
                    return normalize_string(value)
        except Exception:
            pass
    return ""


def users_are_equal(user_from_spreadsheet: dict, user_from_portal: dict, indent: int = 0):
    user_from_spreadsheet = deepcopy(user_from_spreadsheet)
    user_from_spreadsheet[PROPERTY.EMAIL] = user_from_spreadsheet[PROPERTY.EMAIL].lower()
    user_from_portal = deepcopy(user_from_portal)
    user_from_portal[PROPERTY.EMAIL] = user_from_portal[PROPERTY.EMAIL].lower()
    del user_from_portal["uuid"]
    return user_from_spreadsheet == user_from_portal


def compile_diffs(user_from_spreadsheet: dict, user_from_portal: dict) -> List[str]:
    diffs = []
    if isinstance(user_from_spreadsheet, dict) and isinstance(user_from_portal, dict):
        for key in user_from_spreadsheet:
            spreadsheet_value = user_from_spreadsheet[key]
            portal_value = user_from_portal.get(key)
            if portal_value is None:
                diffs.append(f"MISSING: {key} {chars.dot} SPREADSHEET: {spreadsheet_value}")
            elif portal_value != spreadsheet_value:
                diffs.append(f"DIFF: {key} {chars.dot} SPREADSHEET: {spreadsheet_value}"
                             f" {chars.dot} PORTAL: {portal_value}")
        for key in user_from_portal:
            if key == "uuid":
                continue
            portal_value = user_from_portal[key]
            spreadsheet_value = user_from_spreadsheet.get(key)
            if spreadsheet_value is None:
                diffs.append(f"MISSING: {key} {chars.dot} PORTAL: {portal_value}")
    return diffs


@lru_cache(maxsize=1)
def portal_get_users(portal: Portal) -> List[dict]:
    users_query = f"/users?status=deleted&status=current&status=inactive&status=revoked&limit=100000"
    return portal.get(users_query).json().get("@graph")


def portal_get_user(portal: Portal, email: str) -> Optional[dict]:
    if isinstance(email, str) and email:
        for user in portal_get_users(portal):
            if user.get("email") == email:
                return user
    return None


def normalize_string(value: str) -> str:
    if isinstance(value, str):
        # Gets rid of ISO-8859/European characters.
        return "".join([c for c in unicodedata.normalize("NFD", value) if not unicodedata.combining(c)]).strip()
    return ""


def _debug(message: str) -> None:
    print("DEBUG: " + message, file=sys.stderr, flush=True)


def _info(message: str) -> None:
    print(message, file=sys.stderr, flush=True)


def _error(message: str, plain: bool = False) -> None:
    print(("ERROR: " if plain is not True else "") + message, file=sys.stderr, flush=True)


# https://docs.google.com/spreadsheets/d/1WYc_UKkpVUUBPJPbZdwwBX28f6Qr5I9PblOdGqu7g1M/view?gid=446162554#gid=446162554
users_spreadsheet = os.path.expanduser("~/tmp/smaht_users_from_spreadsheet_20241126.tsv")

portal_env = "smaht-data"
dump = False
process = True
verbose = True
debug = False

users = read_users_from_spreadsheet(users_spreadsheet)

if dump:
    print(json.dumps(users, indent=4))

if process:
    portal = Portal(portal_env)
    users_to_add, users_to_update = \
        sanity_check_users_from_spreadsheet_with_portal(users, portal, verbose=verbose, debug=debug)

if users_to_add:
    print("USERS TO ADD")
    print(json.dumps(users_to_add, indent=4))

if users_to_update:
    print("USERS TO UPDATE")
    print(json.dumps(users_to_update, indent=4))



from functools import lru_cache
import boto3
import sys
from typing import Optional, List
from hms_utils.dictionary_utils import sort_dictionary
from dcicutils.misc_utils import format_size
from dcicutils.portal_utils import Portal


@lru_cache(maxsize=1024)
def get_submission_center_info(submission_center_path: str) -> Optional[dict]:
    try:
        return portal.get(submission_center_path).json()
    except Exception:
        return {}


@lru_cache(maxsize=1024)
def get_submission_center_name(submission_center_path: str) -> str:
    return get_submission_center_info(submission_center_path).get("identifier", "")


def get_submission_centers(portal: Portal, submission_centers: List[dict]) -> List[str]:
    submission_center_names = []
    for submission_center in submission_centers:
        if submission_center_path := submission_center.get("@id"):
            if submission_center_name := get_submission_center_name(submission_center_path):
                if submission_center_name not in submission_center_names:
                    submission_center_names.append(submission_center_name)
    return submission_center_names


@lru_cache(maxsize=1024)
def get_submission_center_display_name(submission_center: str) -> str:
    try:
        return portal.get(f"/submission-centers/{submission_center}").json()["display_title"]
    except Exception:
        return ""


def get_file_size(file: dict) -> Optional[int]:
    if isinstance(file, dict) and (key := file.get("upload_key")) and (file.get("status") != "uploading"):
        bucket = "smaht-production-application-files"
        try:
            return boto3.client("s3").head_object(Bucket=bucket, Key=key)["ContentLength"]
        except Exception:
            pass
    return None


portal = Portal("smaht-data")
total = None ; limit = 220 ; offset = 0 ; nfiles_total = 0  # noqa

file_info_per_submission_center = {}

while True:
    url = f"/submitted-files?sort=uuid&from={offset}&limit={limit}"
    files = portal.get(url).json()
    files = files["@graph"]
    nfiles = len(files)
    for file in files:
        submission_centers = get_submission_centers(portal, file.get("submission_centers"))
        if (file_size := file.get("file_size")) is None:
            file_size = get_file_size(file)
        file_info = {
            "uuid": file.get("uuid"),
            "name": file.get("filename"),
            "size": file_size,
            "status": file.get("status"),
            "created": file.get("date_created"),
            "submission_centers": submission_centers
        }
        if not submission_centers:
            submission_center = None
        elif len(submission_centers) > 1:
            print(f"WARNING: More than one submission-center for submitted file:"
                  f" {file_info['name']} | {file_info['uuid']} | {', '.join(submission_centers)}",
                  file=sys.stderr, flush=True)
            submission_center = submission_centers[0]
        else:
            submission_center = submission_centers[0]
        if not (submission_center_file_info := file_info_per_submission_center.get(submission_center)):
            file_info_per_submission_center[submission_center] = [file_info]
        else:
            submission_center_file_info.append(file_info)
        print(f"Submitted Files: {nfiles_total}\r", end="", file=sys.stderr, flush=True)
        nfiles_total += 1
        if (total is not None) and (nfiles_total >= total):
            break
    if ((total is not None) and (nfiles_total >= total)) or (nfiles < limit):
        break
    offset += nfiles

file_summary_per_submission_center = {}
for submission_center in file_info_per_submission_center:
    total_files = 0 ; total_size = 0  # noqa
    for item in file_info_per_submission_center[submission_center]:
        total_files += 1
        total_size += (size if isinstance(size := item["size"], int) else 0)
    file_summary_per_submission_center[submission_center] = {"total_files": total_files, "total_size": total_size}
file_summary_per_submission_center = sort_dictionary(file_summary_per_submission_center)

print(f"Total submitted files and sizes by submission center:")
for submission_center in file_summary_per_submission_center:
    summary = file_summary_per_submission_center[submission_center]
    total_files = summary["total_files"]
    total_size = summary["total_size"]
    # print(f"{submission_center}: {total_files} | {format_size(total_size)}")
    print(f"{get_submission_center_display_name(submission_center)} ({submission_center}):")
    print(f"- Total Files: {total_files}")
    print(f"- Total Size: {format_size(total_size)} ({total_size})")

total_files = sum(item["total_files"] for item in file_summary_per_submission_center.values())
total_size = sum(item["total_size"] for item in file_summary_per_submission_center.values())
print("TOTAL:")
print(f"- Total Files: {total_files}")
print(f"- Total Size: {format_size(total_size)} ({total_size})")



from copy import deepcopy
from functools import lru_cache
import json
from prettytable import PrettyTable
import sys
from typing import Any, Callable, List, Literal, Optional, Union
from hms_utils.argv import ARGV
from hms_utils.chars import chars
from hms_utils.portal.portal_utils import Portal
from hms_utils.type_utils import is_uuid


class Action:
    ADD = "add"
    EDIT = "edit"
    CREATE = "create"
    EXPAND = "expand"
    INDEX = "index"
    LIST = "list"
    RESTRICTED_FIELDS = "restricted_fields"
    SEARCH = "search"
    VIEW = "view"
    VIEW_DETAILS = "view_details"
    VIEW_RAW = "view_raw"
    VISIBLE_FOR_EDIT = "visible_for_edit"


ActionType = Literal[
    Action.ADD,
    Action.EDIT,
    Action.CREATE,
    Action.EXPAND,
    Action.INDEX,
    Action.LIST,
    Action.RESTRICTED_FIELDS,
    Action.SEARCH,
    Action.VIEW,
    Action.VIEW_DETAILS,
    Action.VIEW_RAW,
    Action.VISIBLE_FOR_EDIT
]


def main():

    argv = ARGV({
        ARGV.OPTIONAL(str): ["item"],
        ARGV.OPTIONAL(str): ["--env"],
        ARGV.OPTIONAL(str): ["--user"],
        ARGV.OPTIONAL([str]): ["--actions", "--action"],
        ARGV.OPTIONAL(bool): ["--ping"],
        ARGV.OPTIONAL(bool): ["--verbose"],
        ARGV.OPTIONAL(bool): ["--debug"],
        ARGV.OPTIONAL(bool): ["--version"]
    })

    status = 0

    if not (portal := Portal.create(argv.env, verbose=argv.verbose, debug=argv.debug, ping=argv.ping)):
        error()

    if argv.user:
        if not (portal_for_user := Portal.create(argv.user or argv.env, verbose=argv.verbose, debug=argv.debug)):
            error()
        if portal.server != portal_for_user.server:
            error(f"Mismatched Portal environments: {portal.server} vs {portal_for_user.server}")
    else:
        portal_for_user = portal

    if not (user := portal_for_user.get_metadata("/me", raise_exception=False)):
        error("Cannot retrieve user info from Portal")

    user_email = user.get("email")
    user_uuid = user.get("uuid")
    user_principals = get_user_principals(portal_for_user)

    if argv.item:
        if not (principals_allowed_for_item := get_principals_allowed_for_item(portal, argv.item)):
            error(f"Cannot access Portal item: {argv.item}")
    else:
        principals_allowed_for_item = None

    print(f"\n{chars.rarrow} USER: {user_email} {chars.dot} {user_uuid}")
    print_user_principals(user_principals, principals_allowed_for_item, portal)
    if argv.debug:
        print(json.dumps(user_principals, indent=4))

    if principals_allowed_for_item:

        item_type = get_portal_item_type(portal, argv.item)
        print(f"\n{chars.rarrow} PORTAL ITEM: {argv.item}{f' {chars.dot} {item_type}' if item_type else ''}")
        print_principals_with_actions(principals_allowed_for_item, user_principals)
        if argv.debug:
            print(json.dumps(principals_allowed_for_item, indent=4))

        if (actions := argv.actions) or (actions := sorted(list(principals_allowed_for_item.keys()))):
            print(f"\n{chars.rarrow} PERMISSIONS: {argv.item} {chars.larrow_hollow} {user_email}")
            for action in actions:
                action_allowed = is_user_allowed_item_access(user_principals, principals_allowed_for_item, action)
                print(f"  {chars.dot} ACTION: {action} {chars.rarrow_hollow}"
                      f" {f'ALLOWED' if action_allowed else 'FORBIDDEN'}")
                if not action_allowed:
                    status = 1
    print()

    return status


def get_user_principals(portal_or_key_or_key_name: Union[Portal, dict, str]) -> List[str]:
    """
    Returns the principals for the given user, specified be either a Portal object,
    or a key name (i.e. e.g. in the ~/.smaht-keys.json file), or a key (i.e. e.g. a
    dictionary like {"key": "REDACTED", "secret", server: "http://data.smaht.org"}).
    The result looks something like this:
        [
            "system.Everyone",
            "system.Authenticated",
            "role.submission_center_member_rw.9626d82e-8110-4213-ac75-0a50adf890ff",
            "accesskey.3HJDFQAD",
            "submits_for.9626d82e-8110-4213-ac75-0a50adf890ff",
            "role.consortium_member_rw",
            "group.submitter",
            "role.consortium_member_create",
            "userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68",
            "role.consortium_member_rw.358aed10-9b9d-4e26-ab84-4bd162da182b"
        ]
    Note that this uses the (circa 2024-11-14) debugging/troubleshooting endpoint /debug_user_principals.
    """
    if isinstance(portal_or_key_or_key_name, Portal):
        portal_for_user = portal_or_key_or_key_name
    elif isinstance(portal_or_key_or_key_name, (dict, str)):
        portal_for_user = Portal(portal_or_key_or_key_name)
    else:
        return []
    if not (user_principals := portal_for_user.get_metadata("/debug_user_principals")):
        return []
    return user_principals


def get_principals_allowed_for_item(portal: Portal, query: str,
                                    action: Optional[ActionType] = None) -> Union[dict, List[str]]:
    """
    Returns the principals allowed fro the given Portal item (query). If an action (e.g. view or edit)
    is NOT given then the result looks something like this:
        {
            "view": [
                "group.admin",
                "group.read-only-admin",
                "remoteuser.INDEXER",
                "role.submission_center_member_rw",
                "role.submission_center_member_rw.9626d82e-8110-4213-ac75-0a50adf890ff"
            ],
            "edit": [
                "group.admin",
                "group.submitter"
            ]
        }
    If an action IS given (e.g. view or edit) then the result looks something like this:
        [
            "group.admin",
            "group.read-only-admin",
            "remoteuser.INDEXER",
            "role.submission_center_member_rw",
            "role.submission_center_member_rw.9626d82e-8110-4213-ac75-0a50adf890ff"
        ]
    """
    if action not in ActionType:
        action = None
    if not (isinstance(portal, Portal) and isinstance(query, str) and query):
        return [] if action else {}
    # if not isinstance(item := portal.get_metadata(query, raise_exception=False), dict):
    if not isinstance(item := get_portal_item(portal, query), dict):
        return [] if action else {}
    if not isinstance(principals_allowed_for_item := item.get("principals_allowed"), dict):
        return [] if action else {}
    if action:
        if not isinstance(principals_allowed_for_item_for_action := principals_allowed_for_item.get(action), list):
            return []
        return principals_allowed_for_item_for_action
    else:
        return principals_allowed_for_item


@lru_cache(maxsize=1)
def get_portal_item(portal: Portal, query: str) -> Optional[dict]:
    return portal.get_metadata(query, raise_exception=False)


def get_portal_item_type(portal: Portal, query: str) -> Optional[dict]:
    if item := get_portal_item(portal, query):
        return Portal.get_item_type(item)
    return None


def is_user_allowed_item_access(user_principals_or_portal_or_key_or_key_name: Union[List[str], Portal, dict, str],
                                principals_allowed_for_item: dict, action: str) -> bool:
    if isinstance(user_principals_or_portal_or_key_or_key_name, (Portal, dict, str)):
        user_principals = get_user_principals(user_principals_or_portal_or_key_or_key_name)
    elif isinstance(user_principals_or_portal_or_key_or_key_name, list):
        user_principals = user_principals_or_portal_or_key_or_key_name
    else:
        return False
    if not isinstance(user_principals, list):
        return False
    if not isinstance(principals_allowed_for_item, dict):
        return False
    if not (isinstance(action, str) and (action := action.strip())):
        return False
    if not isinstance(principals_allowed_for_item_for_action := principals_allowed_for_item.get(action), list):
        return False
    for principal_allowed_for_item_for_action in principals_allowed_for_item_for_action:
        if principal_allowed_for_item_for_action in user_principals:
            return True
    return False


def print_principals(principals: List[str], message: Optional[str] = None,
                     value_callback: Optional[Callable] = None,
                     value_header: Optional[str] = None,
                     principal_callback: Optional[Callable] = None,
                     principal_uuid_callback: Optional[Callable] = None,
                     parse_userid: bool = False,
                     header: Optional[List[str]] = None,
                     noheader: bool = False,
                     return_value: bool = False,
                     nosort: bool = False) -> Optional[str]:

    def uuid_specific_principal(principal):  # noqa
        principal_uuid = ""
        if principal.startswith("role."):
            principal = principal.replace("role.", "role/")
        if (index := principal.find(".")) > 0:
            if is_uuid(value := principal[index + 1:]):
                principal_uuid = value
                principal = principal[0:index]
        if principal.startswith("role/"):
            principal = principal.replace("role/", "role.")
        return principal, principal_uuid

    header = header if isinstance(header, list) and (noheader is not True) else None
    rows = []
    for principal in principals:
        principal, principal_uuid = uuid_specific_principal(principal)
        if callable(principal_callback):
            if principal_annotation := principal_callback(principal, principal_uuid):
                principal += principal_annotation
        if callable(principal_uuid_callback):
            if principal_uuid_annotation := principal_uuid_callback(principal_uuid):
                principal_uuid += f" {chars.dot} {principal_uuid_annotation}"
        if principal_uuid:
            rows.append([principal, principal_uuid])
            if value_callback:
                rows[len(rows) - 1].append("")
        else:
            rows.append([principal, ""])
            if value_callback:
                rows[len(rows) - 1].append("")

    table = PrettyTable()
    if header:
        table.field_names = header
    elif noheader is not True:
        if callable(value_callback):
            table.field_names = ["PRINCIPAL/ROLE", "ASSOCIATED UUID",
                                 value_header if isinstance(value_header, str) else ""]
        else:
            table.field_names = ["PRINCIPAL/ROLE", "ASSOCIATED UUID"]
    else:
        table.header = False
    for row in rows:
        if value_callback and (role := value_callback(row[0], row[1])):
            row[2] = role
    if nosort is not True:
        rows.sort(key=lambda row: f"{row[0]}_{row[1]}")
    for row in rows:
        table.add_row([item or chars.null for item in row])
    table.align = "l"
    output = f"{message}\n" if message else ""
    output += str(table)
    if return_value:
        return output
    print(output)


def print_user_principals(user_principals: List[str],
                          principals_allowed_for_item: Optional[dict] = None,
                          portal: Optional[Portal] = None) -> None:
    if isinstance(principals_allowed_for_item, dict):
        principals_allowed_for_item_flattened = [item for items in principals_allowed_for_item.values()
                                                 for item in items]
        principals_allowed_for_item_flattened = list(set(principals_allowed_for_item_flattened))
    else:
        principals_allowed_for_item_flattened = None
    def principal_annotation(principal: str, principal_uuid: Optional[str] = None):  # noqa
        nonlocal principals_allowed_for_item_flattened
        if isinstance(principal, str) and principal:
            if isinstance(principal_uuid, str) and principal_uuid:
                principal += f"{principal_uuid}"
            if principal in principals_allowed_for_item_flattened:
                return f" {chars.larrow}{chars.larrow}{chars.larrow}"
    def principal_uuid_annotation(principal_uuid: str) -> Optional[str]:  # noqa
        nonlocal portal
        return get_affiliation(portal, principal_uuid)
    print_principals(user_principals,
                     principal_callback=principal_annotation if principals_allowed_for_item_flattened else None,
                     principal_uuid_callback=principal_uuid_annotation if isinstance(portal, Portal) else None)


def print_principals_with_actions(principals_allowed_for_item_by_action: dict,
                                  user_principals: Optional[List[str]] = None) -> None:
    def find_actions(principal, principal_uuid) -> Optional[str]:
        nonlocal principals_allowed_for_item_by_action
        if isinstance(principal, str):
            if is_uuid(principal_uuid):
                principal += f".{principal_uuid}"
            actions = [action for action, items in principals_allowed_for_item_by_action.items() if principal in items]
            return f" {chars.dot} ".join(actions)
    def principal_annotation(principal: str, principal_uuid: Optional[str] = None) -> None:  # noqa
        nonlocal user_principals
        if isinstance(principal, str) and principal:
            if isinstance(principal_uuid, str) and principal_uuid:
                principal += f"{principal_uuid}"
            if principal in user_principals:
                return f" {chars.larrow}{chars.larrow}{chars.larrow}"
    principals_allowed_for_item = [item for items in principals_allowed_for_item_by_action.values() for item in items]
    principals_allowed_for_item = list(set(principals_allowed_for_item))
    print_principals(principals_allowed_for_item,
                     principal_callback=principal_annotation if isinstance(user_principals, list) else None,
                     value_callback=find_actions, value_header="ACTION     ")


def print_acls(acls: List[tuple], message: Optional[str] = None, nosort: bool = False) -> None:
    rows = []
    for acl_item in acls:
        acl_permission, acl_principal, acl_actions = acl_item
        if isinstance(acl_actions, list):
            for index in range(len(acl_actions)):
                if str(acl_actions[index]).startswith("<pyramid.security.AllPermissionsList"):
                    acl_actions[index] = "pyramid.ALL_PERMISSIONS"
        elif str(acl_actions).startswith("<pyramid.security.AllPermissionsList"):
            acl_actions = "pyramid.ALL_PERMISSIONS"
        if isinstance(acl_actions, list):
            acl_actions = f" {chars.dot} ".join(acl_actions)
        rows.append([acl_principal, f"{acl_permission} {chars.larrow_hollow} {acl_actions}"])
    table = PrettyTable()
    table.field_names = ["ACL PRINCIPAL/ROLE", "ACL PERMISSION/ACTION"]
    if nosort is not True:
        rows.sort(key=lambda row: row[0])
    for row in rows:
        table.add_row(row)
    table.align = "l"
    if message:
        print(message)
    print(table)


def print_principals_with_acls(principals: List[str], acls: List[tuple],
                               action: Optional[Any] = None,
                               message: Optional[str] = None) -> None:
    def find_matching_acls() -> Optional[tuple]:  # noqa
        nonlocal principals, acls, action
        matching_acls = []
        if action:
            for acl_item in acls:
                acl_permission, acl_principal, acl_actions = acl_item
                if (acl_principal in principals) and (action in acl_actions):
                    matching_acls.append(acl_item)
        return matching_acls
    def action_allowed() -> bool:  # noqa
        nonlocal principals, acls, action
        if action:
            for acl_item in acls:
                acl_permission, acl_principal, acl_actions = acl_item
                if (acl_principal in principals) and (action in acl_actions):
                    return acl_permission == "Allow"
        return False
    matching_acls = find_matching_acls()
    def acl_value(principal, principal_uuid):  # noqa
        nonlocal acls, principals, action
        if principal_uuid:
            principal = f"{principal}.{principal_uuid}"
        values = []
        for acl_item in acls:
            acl_permission, acl_principal, acl_actions = acl_item
            if acl_principal == principal:
                acl_actions_display = deepcopy(acl_actions)
                if isinstance(acl_actions_display, list):
                    for index in range(len(acl_actions_display)):
                        if str(acl_actions_display[index]).startswith("<pyramid.security.AllPermissionsList"):
                            acl_actions_display[index] = "pyramid.ALL_PERMISSIONS"
                elif str(acl_actions_display).startswith("<pyramid.security.AllPermissionsList"):
                    acl_actions_display = "pyramid.ALL_PERMISSIONS"
                if isinstance(acl_actions_display, list):
                    acl_actions_display = f" {chars.dot} ".join(acl_actions_display)
                value = f"{acl_permission} {chars.larrow_hollow} {acl_actions_display}"
                if action:
                    if acl_item in matching_acls:
                        value += f" {chars.larrow}{chars.larrow}{chars.larrow}"
                        if acl_permission == "Allow":
                            value += f" {chars.check}"
                        elif acl_permission == "Deny":
                            value += f" {chars.xmark}"
                    elif action in acl_actions:
                        value += f" {chars.larrow}"
                values.append(value)
        return "\n".join(values)
    if action:
        value_header = f"ACL {chars.larrow_hollow} {action}"
        if action_allowed():
            value_header += f" {chars.check}{chars.check}{chars.check}"
        else:
            value_header += f" {chars.xmark}"
    else:
        value_header = "ACL"
    print_principals(principals, value_callback=acl_value, value_header=value_header, message=message)


def get_affiliation(portal: Portal, uuid: str) -> Optional[str]:
    if isinstance(portal, Portal) and is_uuid(uuid):
        if isinstance(item := portal.get_metadata(uuid, raise_exception=True), dict):
            return item.get("identifier")
    return None


def error(message: Optional[str] = None, exit: bool = True, status: int = 1) -> None:
    if isinstance(message, str) and message:
        print(f"ERROR: {str(message)}", file=sys.stderr)
    if exit is not False:
        sys.exit(status if isinstance(status, int) else 1)


if __name__ == "__main__":
    sys.exit(main())


if False:
    principals = \
        {"group.admin", 'role.consortium_member_rw', 'role.consortium_member_create', 'submits_for.9626d82e-8110-4213-ac75-0a50adf890ff', 'group.submitter', 'accesskey.POOZOKLE', 'system.Authenticated', 'role.consortium_member_rw.358aed10-9b9d-4e26-ab84-4bd162da182b', 'role.submission_center_member_rw', 'role.submission_center_member_create', 'userid.0d565156-00db-4948-82fe-021d151a5daf', 'role.submission_center_member_rw.9626d82e-8110-4213-ac75-0a50adf890ff', 'system.Everyone'}  # noqa

    acls = \
        [('Allow', 'system.Everyone', ['list', 'search']), ('Allow', 'group.admin', '<pyramid.security.AllPermissionsList object at 0x12724a330>'), ('Allow', 'remoteuser.EMBED', 'restricted_fields'), ('Allow', 'remoteuser.INDEXER', ['view', 'view_raw', 'list', 'index']), ('Allow', 'remoteuser.EMBED', ['view', 'view_raw', 'expand']), ('Allow', 'system.Everyone', ['visible_for_edit'])]  # noqa

    print_principals(principals)
    print_acls(acls)
    action = "visible_for_edit"
    action = "search"
    print_principals_with_acls(principals, acls, action=action)



# ------------------------------------------------------------------------------------------------------
# Command-line utility to retrieve and print the given object (UUID) from a SMaHT/CGAP/Fourfront Portal.
# ------------------------------------------------------------------------------------------------------
# Example command:
#  view-portal-object 4483b19d-62e7-4e7f-a211-0395343a35df --yaml
#
# Example output:
#  '@context': /terms/
#  '@id': /access-keys/3968e38e-c11f-472e-8531-8650e2e296d4/
#  '@type':
#  - AccessKey
#  - Item
#  access_key_id: NSVCZ75O
#  date_created: '2023-09-06T13:11:59.704005+00:00'
#  description: Manually generated local access-key for testing.
#  display_title: AccessKey from 2023-09-06
#  expiration_date: '2023-12-05T13:11:59.714106'
#  last_modified:
#    date_modified: '2023-09-06T13:11:59.711367+00:00'
#    modified_by:
#      '@id': /users/3202fd57-44d2-44fb-a131-afb1e43d8ae5/
#      '@type':
#      - User
#      - Item
#      status: current
#      uuid: 3202fd57-44d2-44fb-a131-afb1e43d8ae5
#  principals_allowed:
#    edit:
#    - group.admin
#    - userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68
#    view:
#    - group.admin
#    - group.read-only-admin
#    - userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68
#  schema_version: '1'
#  status: current
#  user:
#    '@id': /users/74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68/
#    '@type':
#    - User
#    - Item
#    display_title: David Michaels
#    principals_allowed:
#      edit:
#      - group.admin
#      view:
#      - group.admin
#      - group.read-only-admin
#    status: current
#    uuid: 74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68
#  uuid: 3968e38e-c11f-472e-8531-8650e2e296d4
#
# Note that instead of a uuid you can also actually use a path, for example:
#   view-local-object /file-formats/vcf_gz_tbi
#
# --------------------------------------------------------------------------------------------------

import argparse
from functools import lru_cache
import io
import json
import pyperclip
import os
import sys
from typing import Callable, List, Optional, TextIO, Tuple, Union
import yaml
from dcicutils.captured_output import captured_output, uncaptured_output
from dcicutils.command_utils import yes_or_no
from dcicutils.misc_utils import get_error_message, PRINT, to_snake_case
from dcicutils.portal_utils import Portal
from hms_utils.dictionary_utils import sort_dictionary
from hms_utils.threading_utils import run_concurrently
from hms_utils.type_utils import is_uuid


# Schema properties to ignore (by default) for the view schema usage.
_SCHEMAS_IGNORE_PROPERTIES = [
    "date_created",
    "last_modified",
    "principals_allowed",
    "submitted_by",
    "schema_version"
]

_output_file: TextIO = None


def main():

    global _output_file

    parser = argparse.ArgumentParser(description="View Portal object.")
    parser.add_argument("uuid", nargs="?", type=str,
                        help=f"The uuid (or path) of the object to fetch and view. ")
    parser.add_argument("--ini", type=str, required=False, default=None,
                        help=f"Name of the application .ini file.")
    parser.add_argument("--env", "-e", type=str, required=False, default=None,
                        help=f"Environment name (key from ~/.smaht-keys.json).")
    parser.add_argument("--server", "-s", type=str, required=False, default=None,
                        help=f"Environment server name (server from key in ~/.smaht-keys.json).")
    parser.add_argument("--app", type=str, required=False, default=None,
                        help=f"Application name (one of: smaht, cgap, fourfront).")
    parser.add_argument("--schema", action="store_true", required=False, default=False,
                        help="View named schema rather than object.")
    parser.add_argument("--all", action="store_true", required=False, default=False,
                        help="Include all properties for schema usage.")
    parser.add_argument("--raw", action="store_true", required=False, default=False, help="Raw output.")
    parser.add_argument("--inserts", action="store_true", required=False, default=False,
                        help="Format output for subsequent inserts.")
    parser.add_argument("--insert-files", action="store_true", required=False, default=False,
                        help="Output for to insert files.")
    parser.add_argument("--ignore", nargs="+", help="Ignore these fields for --inserts.")
    parser.add_argument("--tree", action="store_true", required=False, default=False, help="Tree output for schemas.")
    parser.add_argument("--database", action="store_true", required=False, default=False,
                        help="Read from database output.")
    parser.add_argument("--check", "--bool", action="store_true", required=False,
                        default=False, help="Only return whether found or not.")
    parser.add_argument("--refs", action="store_true", required=False,
                        default=False, help="Include object for referenced uuids.")
    parser.add_argument("--yaml", action="store_true", required=False, default=False, help="YAML output.")
    parser.add_argument("--copy", "-c", action="store_true", required=False, default=False,
                        help="Copy object data to clipboard.")
    parser.add_argument("--output", required=False, help="Output file.", type=str)
    parser.add_argument("--indent", required=False, default=False, help="Indent output.", type=int)
    parser.add_argument("--summary", action="store_true", required=False, default=False,
                        help="Summary output (for schema only).")
    parser.add_argument("--force", action="store_true", required=False, default=False, help="Debugging output.")
    parser.add_argument("--sort", action="store_true", required=False, default=False, help="Sort output.")
    parser.add_argument("--terse", action="store_true", required=False, default=False, help="Terse output.")
    parser.add_argument("--verbose", action="store_true", required=False, default=False, help="Verbose output.")
    parser.add_argument("--noheader", action="store_true", required=False, default=False, help="Supress header output.")
    parser.add_argument("--debug", action="store_true", required=False, default=False, help="Debugging output.")
    args = parser.parse_args()

    portal = _create_portal(ini=args.ini, env=args.env or os.environ.get("SMAHT_ENV"),
                            server=args.server, app=args.app,
                            verbose=args.verbose and not args.noheader, debug=args.debug)

    if not args.uuid:
        _print("UUID or schema or path required.")
        _exit(1)

    if args.insert_files:
        args.inserts = True
        if args.output:
            if not os.path.isdir(args.output):
                _print(f"Specified output directory for insert files does not exist: {args.output}")
                exit(1)
            args.insert_files = args.output
            args.output = None

    if args.output:
        if os.path.exists(args.output):
            if os.path.isdir(args.output):
                _print(f"Specified output file already exists as a directory: {args.output}")
                _exit(1)
            elif os.path.isfile(args.output):
                _print(f"Specified output file already exists: {args.output}")
                if (not args.force) and not yes_or_no(f"Do you want to overwrite this file?"):
                    _exit(0)
        _output_file = io.open(args.output, "w")

    if args.uuid and ((args.uuid.lower() == "schemas") or (args.uuid.lower() == "schema")):
        _print_all_schema_names(portal=portal, terse=args.terse, all=args.all,
                                tree=args.tree, summary=args.summary, yaml=args.yaml)
        return
    elif args.uuid and (args.uuid.lower() == "info"):
        if consortia := portal.get_metadata("/consortia?limit=1000"):
            _print_output("Known Consortia:")
            consortia = sorted(consortia.get("@graph", []), key=lambda key: key.get("identifier"))
            for consortium in consortia:
                if ((consortium_name := consortium.get("identifier")) and
                    (consortium_uuid := consortium.get("uuid"))):  # noqa
                    _print_output(f"- {consortium_name}: {consortium_uuid}")
        if submission_centers := portal.get_metadata("/submission-centers?limit=1000"):
            _print_output("Known Submission Centers:")
            submission_centers = sorted(submission_centers.get("@graph", []), key=lambda key: key.get("identifier"))
            for submission_center in submission_centers:
                if ((submission_center_name := submission_center.get("identifier")) and
                    (submission_center_uuid := submission_center.get("uuid"))):  # noqa
                    _print_output(f"- {submission_center_name}: {submission_center_uuid}")
        try:
            if file_formats := portal.get_metadata("/file-formats?limit=1000"):
                _print_output("Known File Formats:")
                file_formats = sorted(file_formats.get("@graph", []), key=lambda key: key.get("identifier"))
                for file_format in file_formats:
                    if ((file_format_name := file_format.get("identifier")) and
                        (file_format_uuid := file_format.get("uuid"))):  # noqa
                        _print_output(f"- {file_format_name}: {file_format_uuid}")
        except Exception:
            _print_output("Known File Formats: None")
        return

    if _is_maybe_schema_name(args.uuid):
        args.schema = True

    if args.schema:
        schema, schema_name = _get_schema(portal, args.uuid)
        if schema:
            if args.copy:
                pyperclip.copy(json.dumps(schema, indent=4))
            if args.summary:
                if parent_schema_name := _get_parent_schema_name(schema):
                    if schema.get("isAbstract") is True:
                        _print_output(f"{schema_name} | parent: {parent_schema_name} | abstract")
                    else:
                        _print_output(f"{schema_name} | parent: {parent_schema_name}")
                else:
                    _print_output(schema_name)
            _print_schema(schema, terse=args.terse,
                          all=args.all, summary=args.summary, yaml=args.yaml)
            return

    data = _get_portal_object(portal=portal, uuid=args.uuid, raw=args.raw, database=args.database,
                              inserts=args.inserts, insert_files=args.insert_files,
                              ignore=args.ignore, check=args.check,
                              force=args.force, verbose=args.verbose, debug=args.debug)
    if args.insert_files:
        return

    if args.check:
        if data:
            _print(f"{args.uuid}: found")
            _exit(0)
        else:
            _print(f"{args.uuid}: not found")
            _exit(1)
    if args.copy:
        pyperclip.copy(json.dumps(data, indent=4))
    if args.yaml:
        _print_output(yaml.dump(data))
    else:
        if args.refs and args.inserts:
            fetched_referenced_uuids = []
            while True:
                if args.debug:
                    print(f"Looking for referenced items.")
                if not (referenced_uuids := get_referenced_uuids(data, fetched_referenced_uuids)):
                    if args.debug:
                        print(f"No more referenced items.")
                    break
                fetch_reference_functions = []
                fetch_reference_function_results = []
                for referenced_uuid in referenced_uuids:
                    if args.debug:
                        print(f"Fetching referenced item: {referenced_uuid}")
                    def fetch_reference(referenced_uuid):  # noqa
                        nonlocal portal, fetched_referenced_uuids
                        if referenced_uuid in fetched_referenced_uuids:
                            return
                        fetch_reference_function_results.append(
                            _get_portal_object(portal=portal, uuid=referenced_uuid,
                                               raw=args.raw, database=args.database,
                                               inserts=args.inserts, insert_files=args.insert_files,
                                               ignore=args.ignore, check=args.check, force=args.force,
                                               verbose=args.verbose, debug=args.debug))
                        fetched_referenced_uuids.append(referenced_uuid)
                    fetch_reference_functions.append(
                        lambda referenced_uuid=referenced_uuid: fetch_reference(referenced_uuid))
                run_concurrently(fetch_reference_functions)
                for referenced_data in fetch_reference_function_results:
                    add_referenced_data_to_data(referenced_data, data)
        if args.sort:
            data = sort_dictionary(data)
        if args.indent > 0:
            _print_output(_format_json_with_indent(data, indent=args.indent))
        else:
            _print_output(json.dumps(data, default=str, indent=4))


def get_referenced_uuids(data: dict, fetched_referenced_uuids: List[str]) -> List[str]:
    referenced_uuids = []
    def find_uuids(item: Union[dict, list, str], ignore_uuid: Optional[str] = None):  # noqa
        if isinstance(item, dict):
            for value in item.values():
                find_uuids(value, ignore_uuid=item.get("uuid"))
        elif isinstance(item, list):
            for value in item:
                find_uuids(value)
        elif isinstance(item, str) and is_uuid(item) and (item not in referenced_uuids):
            if item != ignore_uuid:
                if item not in fetched_referenced_uuids:
                    referenced_uuids.append(item)
    find_uuids(data)
    return referenced_uuids


def add_referenced_data_to_data(referenced_data: dict, data: dict) -> None:
    if isinstance(referenced_data, dict) and isinstance(data, dict):
        for referenced_data_type in referenced_data:
            if not isinstance(referenced_data := referenced_data[referenced_data_type], list):
                continue
            for referenced_data_item in referenced_data:
                referenced_data_item_uuid = referenced_data_item.get("uuid")
                if isinstance(data_items := data.get(referenced_data_type), list):
                    data_item_existing = [
                        element for element in data_items
                        if isinstance(element, dict) and (element.get("uuid") == referenced_data_item_uuid)]
                    if not data_item_existing:
                        data_items.append(referenced_data_item)
                else:
                    data[referenced_data_type] = referenced_data


def _obsolete_contains_item(data: dict, uuid: str) -> bool:
    if isinstance(data, dict) and is_uuid(uuid):
        for data_type in data:
            if isinstance(data_items := data[data_type], list):
                for data_item in data_items:
                    if isinstance(data_item, dict) and (data_item.get("uuid") == uuid):
                        return True
    return False


def _format_json_with_indent(value: dict, indent: int = 0) -> Optional[str]:
    if isinstance(value, dict):
        result = json.dumps(value, indent=4)
        if indent > 0:
            result = f"{indent * ' '}{result}"
            result = result.replace("\n", f"\n{indent * ' '}")
        return result


def _create_portal(ini: str, env: Optional[str] = None,
                   server: Optional[str] = None, app: Optional[str] = None,
                   verbose: bool = False, debug: bool = False) -> Portal:
    portal = None
    with captured_output(not debug):
        portal = Portal(env, server=server, app=app) if env or app else Portal(ini)
    if portal:
        if verbose:
            if portal.env:
                _print(f"Portal environment: {portal.env}")
            if portal.keys_file:
                _print(f"Portal keys file: {portal.keys_file}")
            if portal.key_id:
                _print(f"Portal key prefix: {portal.key_id[0:2]}******")
            if portal.ini_file:
                _print(f"Portal ini file: {portal.ini_file}")
            if portal.server:
                _print(f"Portal server: {portal.server}")
        return portal


def _get_portal_object(portal: Portal, uuid: str,
                       raw: bool = False, database: bool = False,
                       inserts: bool = False, insert_files: bool = False,
                       ignore: Optional[List[str]] = None,
                       check: bool = False, force: bool = False,
                       verbose: bool = False, debug: bool = False) -> dict:

    def prune_data(data: dict) -> dict:
        nonlocal ignore
        if not isinstance(ignore, list) or not ignore:
            return data
        return {key: value for key, value in data.items() if key not in ignore}

    def get_metadata_for_individual_result_type(uuid: str) -> Optional[dict]:  # noqa
        # There can be a lot of individual results for which we may need to get the actual type,
        # so do this in a function we were can give verbose output feedback.
        nonlocal portal, results_index, results_total, verbose
        if verbose:
            _print(f"Getting actual type for {results_type} result:"
                   f" {uuid} [{results_index} of {results_total}]", end="")
        result = portal.get_metadata(uuid, raise_exception=False)
        if (isinstance(result_types := result.get("@type"), list) and
            result_types and (result_type := result_types[0])):  # noqa
            if verbose:
                _print(f" -> {result_type}")
            return result_type
        if verbose:
            _print()
        return None

    def get_metadata_type_info(path: str) -> Optional[dict]:
        # Returns a dictionary mapping uuids to their corresponding actual types for the given (search) path.
        nonlocal portal, debug
        metadata_types = {}
        path = f"{path}&field=uuid" if ("?" in path) else f"{path}?field=uuid"
        try:
            if verbose:
                _print(f"Executing separted query to get actual metadata types for raw/inserts query.")
            if ((response := portal.get(path)) and (response.status_code in [200, 307]) and
                (response := response.json()) and (results := response.get("@graph"))):  # noqa
                for result in results:
                    if (result_type := result.get("@type")) and (result_uuid := result.get("uuid")):
                        if ((isinstance(result_type, list) and (result_type := result_type[0])) or
                            isinstance(result_type, str)):  # noqa
                            metadata_types[result_uuid] = result_type
        except Exception:
            return None
        return metadata_types

    def write_insert_files(response: dict) -> None:
        nonlocal insert_files, force
        output_directory = insert_files if isinstance(insert_files, str) else os.getcwd()
        for schema_name in response:
            schema_data = response[schema_name]
            file_name = f"{to_snake_case(schema_name)}.json"
            file_path = os.path.join(output_directory, file_name)
            message_verb = "Writing"
            if os.path.exists(file_path):
                message_verb = "Overwriting"
                if os.path.isdir(file_path):
                    _print(f"WARNING: Output file already exists as a directory. SKIPPING: {file_path}")
                    continue
                if not force:
                    _print(f"Output file already exists: {file_path}")
                    if not yes_or_no(f"Overwrite this file?"):
                        continue
            if verbose:
                _print(f"{message_verb} {schema_name} (object{'s' if len(schema_data) != 1 else ''}:"
                       f" {len(schema_data)}) file: {file_path}")
            with io.open(file_path, "w") as f:
                json.dump(schema_data, f, indent=4)

    if os.path.isfile(uuid) and inserts:
        # Very special case: If given "uuid" (or other path) as actually a file then assume it
        # contains a list of references (e.g. /Donor/3039a6ca-9849-432d-ad49-2c5630bcbee7) to fetch.
        response = {}
        if verbose:
            _print(f"Reading references from file: {uuid}")
        with io.open(uuid) as f:
            for line in f:
                if ((line := line.strip()) and (components := line.split("/")) and (len(components) > 1) and
                    (schema_name := components[1]) and (schema_name := _get_schema(portal, schema_name)[1])):  # noqa
                    try:
                        if ((result := portal.get(line, raw=True, database=database)) and
                            (result.status_code in [200, 307]) and (result := result.json())):  # noqa
                            if not response.get(schema_name):
                                response[schema_name] = []
                            response[schema_name].append(result)
                            continue
                    except Exception:
                        pass
                    _print(f"Cannot get reference: {line}")
            if insert_files:
                write_insert_files(response)
            return response
    else:
        response = None
        try:
            if not uuid.startswith("/"):
                path = f"/{uuid}"
            else:
                path = uuid
            if (response := portal.get(path, raw=raw or inserts, database=database)) is not None:
                if response.status_code == 403:
                    _exit(f"Permission error getting Portal object from {portal.server}: {uuid}")
                elif response.status_code == 404:
                    _exit(f"Not found Portal object from {portal.server}: {uuid}")

        except Exception as e:
            if "404" in str(e) and "not found" in str(e).lower():
                _print(f"Portal object not found at {portal.server}: {uuid}")
                _exit()
            _exit(f"Exception getting Portal object from {portal.server}: {uuid}\n{get_error_message(e)}")
        if not response:
            if check:
                return None
            _exit(f"Null response getting Portal object from {portal.server}: {uuid}")
        if response.status_code not in [200, 307]:
            # TODO: Understand why the /me endpoint returns HTTP status code 307, which is only why we mention it above.
            _exit(f"Invalid status code ({response.status_code}) getting Portal object from {portal.server}: {uuid}")
        if not response.json:
            _exit(f"Invalid JSON getting Portal object: {uuid}")
        response = response.json()

    response_types = {}
    if inserts:
        # Format results as suitable for inserts (e.g. via update-portal-object).
        response.pop("schema_version", None)
        if ((isinstance(results := response.get("@graph"), list) and results) and
            (isinstance(results_type := response.get("@type"), list) and results_type) and
            (isinstance(results_type := results_type[0], str) and results_type.endswith("SearchResults")) and
            (results_type := results_type[0:-len("SearchResults")])):  # noqa
            # For (raw frame) search results, the type (from XyzSearchResults, above) may not be precisely correct
            # for each of the results; it may be the supertype (e.g. QualityMetric vs QualityMetricWorkflowRun);
            # so for types which are supertypes (gotten via Portal.get_schemas_super_type_map) we actually
            # lookup each result individually to determine its actual precise type. Although, if we have
            # more than (say) 5 results to do this for, then do a separate query (get_metadata_type_info)
            # to get the result types all at once.
            if not ((supertypes := portal.get_schemas_super_type_map()) and (subtypes := supertypes.get(results_type))):
                subtypes = None
            response = {}
            results_index = 0
            results_total = len(results)
            for result in results:
                results_index += 1
                if debug:
                    print(f"Processing result: {results_index}")
                result.pop("schema_version", None)
                result = prune_data(result)
                if (subtypes and one_or_more_objects_of_types_exists(portal, subtypes, debug=debug) and
                    (result_uuid := result.get("uuid"))):  # noqa
                    # If we have more than (say) 5 results for which we need to determine that actual result type,
                    # then get them all at once via separate query (get_metadata_type_info)) which is not the raw frame.
                    if (results_total > 5) and (not response_types):
                        response_types = get_metadata_type_info(path)
                    if not (response_types and (result_type := response_types.get(result_uuid))):
                        if individual_result_type := get_metadata_for_individual_result_type(result_uuid):
                            result_type = individual_result_type
                        else:
                            result_type = results_type
                else:
                    result_type = results_type
                if response.get(result_type):
                    response[result_type].append(result)
                else:
                    response[result_type] = [result]
        # Get the result as non-raw so we can get its type.
        elif ((response_cooked := portal.get(path, database=database)) and
              (isinstance(response_type := response_cooked.json().get("@type"), list) and response_type)):
            response = {f"{response_type[0]}": [prune_data(response)]}
        if insert_files:
            write_insert_files(response)
#           output_directory = insert_files if isinstance(insert_files, str) else os.getcwd()
#           for schema_name in response:
#               schema_data = response[schema_name]
#               file_name = f"{to_snake_case(schema_name)}.json"
#               file_path = os.path.join(output_directory, file_name)
#               message_verb = "Writing"
#               if os.path.exists(file_path):
#                   message_verb = "Overwriting"
#                   if os.path.isdir(file_path):
#                       _print(f"WARNING: Output file already exists as a directory. SKIPPING: {file_path}")
#                       continue
#                   if not force:
#                       _print(f"Output file already exists: {file_path}")
#                       if not yes_or_no(f"Overwrite this file?"):
#                           continue
#               if verbose:
#                   _print(f"{message_verb} {schema_name} (object{'s' if len(schema_data) != 1 else ''}:"
#                          f" {len(schema_data)}) file: {file_path}")
#               with io.open(file_path, "w") as f:
#                   json.dump(schema_data, f, indent=4)
    elif raw:
        response.pop("schema_version", None)
    return response


def one_or_more_objects_of_types_exists(portal: Portal, schema_types: List[str], debug: bool = False) -> bool:
    for schema_type in schema_types:
        if one_or_more_objects_of_type_exists(portal, schema_type, debug=debug):
            return True
    return False


@lru_cache(maxsize=64)
def one_or_more_objects_of_type_exists(portal: Portal, schema_type: str, debug: bool = False) -> bool:
    try:
        if debug:
            _print(f"Checking if there are actually any objects of type: {schema_type}")
        if portal.get(f"/{schema_type}").status_code == 404:
            if debug:
                _print(f"No objects of type actually exist: {schema_type}")
            return False
        else:
            if debug:
                _print(f"One or more objects of type exist: {schema_type}")
    except Exception as e:
        _print(f"ERROR: Cannot determine if there are actually any objects of type: {schema_type}")
        _print(e)
    return True


@lru_cache(maxsize=1)
def _get_schemas(portal: Portal) -> Optional[dict]:
    return portal.get_schemas()


def _get_schema(portal: Portal, name: str) -> Tuple[Optional[dict], Optional[str]]:
    if portal and name and (name := name.replace("_", "").replace("-", "").strip().lower()):
        if schemas := _get_schemas(portal):
            for schema_name in schemas:
                if schema_name.replace("_", "").replace("-", "").strip().lower() == name.lower():
                    return schemas[schema_name], schema_name
    return None, None


def _is_maybe_schema_name(value: str) -> bool:
    if value and not is_uuid(value) and not value.startswith("/"):
        return True
    return False


def _is_schema_name(portal: Portal, value: str) -> bool:
    try:
        return _get_schema(portal, value)[0] is not None
    except Exception:
        return False


def _is_schema_named_json_file_name(portal: Portal, value: str) -> bool:
    try:
        return value.endswith(".json") and _is_schema_name(portal, os.path.basename(value[:-5]))
    except Exception:
        return False


def _get_schema_name_from_schema_named_json_file_name(portal: Portal, value: str) -> Optional[str]:
    try:
        if not value.endswith(".json"):
            return None
        _, schema_name = _get_schema(portal, os.path.basename(value[:-5]))
        return schema_name
    except Exception:
        return False


def _print_schema(schema: dict, terse: bool = False, all: bool = False,
                  summary: bool = False, yaml: bool = False) -> None:
    if summary is not True:
        if yaml:
            _print_output(yaml.dump(schema))
        else:
            _print_output(json.dumps(schema, indent=4))
        return
    _print_schema_info(schema, terse=terse, all=all)


def _print_schema_info(schema: dict, level: int = 0,
                       terse: bool = False, all: bool = False,
                       required: Optional[List[str]] = None) -> None:
    if not schema or not isinstance(schema, dict):
        return
    identifying_properties = schema.get("identifyingProperties")
    if level == 0:
        if required_properties := schema.get("required"):
            _print_output("- required properties:")
            for required_property in sorted(list(set(required_properties))):
                if not all and required_property in _SCHEMAS_IGNORE_PROPERTIES:
                    continue
                if property_type := (info := schema.get("properties", {}).get(required_property, {})).get("type"):
                    if property_type == "array" and (array_type := info.get("items", {}).get("type")):
                        _print_output(f"  - {required_property}: {property_type} of {array_type}")
                    else:
                        _print_output(f"  - {required_property}: {property_type}")
                else:
                    _print_output(f"  - {required_property}")
            if isinstance(any_of := schema.get("anyOf"), list):
                if ((any_of == [{"required": ["submission_centers"]}, {"required": ["consortia"]}]) or
                    (any_of == [{"required": ["consortia"]}, {"required": ["submission_centers"]}])):  # noqa
                    # Very very special case.
                    _print_output(f"  - at least one of:")
                    _print_output(f"    - consortia: array of string")
                    _print_output(f"    - submission_centers: array of string")
            required = required_properties
        if identifying_properties := schema.get("identifyingProperties"):
            _print_output("- identifying properties:")
            for identifying_property in sorted(list(set(identifying_properties))):
                if not all and identifying_property in _SCHEMAS_IGNORE_PROPERTIES:
                    continue
                if property_type := (info := schema.get("properties", {}).get(identifying_property, {})).get("type"):
                    if property_type == "array" and (array_type := info.get("items", {}).get("type")):
                        _print_output(f"  - {identifying_property}: {property_type} of {array_type}")
                    else:
                        _print_output(f"  - {identifying_property}: {property_type}")
                else:
                    _print_output(f"  - {identifying_property}")
        if properties := schema.get("properties"):
            reference_properties = []
            for property_name in properties:
                if not all and property_name in _SCHEMAS_IGNORE_PROPERTIES:
                    continue
                property = properties[property_name]
                if link_to := property.get("linkTo"):
                    reference_properties.append({"name": property_name, "ref": link_to})
            if reference_properties:
                _print_output("- reference properties:")
                for reference_property in sorted(reference_properties, key=lambda key: key["name"]):
                    _print_output(f"  - {reference_property['name']}: {reference_property['ref']}")
        if schema.get("additionalProperties") is True:
            _print_output(f"  - additional properties are allowed")
    if terse:
        return
    if properties := (schema.get("properties") if level == 0 else schema):
        if level == 0:
            _print_output("- properties:")
        for property_name in sorted(properties):
            if not all and property_name in _SCHEMAS_IGNORE_PROPERTIES:
                continue
            if property_name.startswith("@"):
                continue
            spaces = f"{' ' * (level + 1) * 2}"
            property = properties[property_name]
            property_required = required and property_name in required
            if property_type := property.get("type"):
                if property_type == "object":
                    suffix = ""
                    if not (object_properties := property.get("properties")):
                        if property.get("additionalProperties") is True:
                            property_type = "any object"
                        else:
                            property_type = "undefined object"
                    elif property.get("additionalProperties") is True:
                        property_type = "open ended object"
                    if property.get("calculatedProperty"):
                        suffix += f" | calculated"
                    _print_output(f"{spaces}- {property_name}: {property_type}{suffix}")
                    _print_schema_info(object_properties, level=level + 1, terse=terse, all=all,
                                       required=property.get("required"))
                elif property_type == "array":
                    suffix = ""
                    if property_required:
                        suffix += f" | required"
                    if property.get("uniqueItems"):
                        suffix += f" | unique"
                    if property.get("calculatedProperty"):
                        suffix += f" | calculated"
                    if property_items := property.get("items"):
                        if (enumeration := property_items.get("enum")) is not None:
                            suffix = f" | enum" + suffix
                        if pattern := property_items.get("pattern"):
                            suffix += f" | pattern: {pattern}"
                        if (format := property_items.get("format")) and (format != "uuid"):
                            suffix += f" | format: {format}"
                        if (max_length := property_items.get("maxLength")) is not None:
                            suffix += f" | max items: {max_length}"
                        if property_type := property_items.get("type"):
                            if property_type == "object":
                                suffix = ""
                                _print_output(f"{spaces}- {property_name}: array of object{suffix}")
                                _print_schema_info(property_items.get("properties"), level=level + 1,
                                                   terse=terse, all=all,
                                                   required=property_items.get("required"))
                            elif property_type == "array":
                                # This (array-of-array) never happens to occur at this time (February 2024).
                                _print_output(f"{spaces}- {property_name}: array of array{suffix}")
                            else:
                                _print_output(f"{spaces}- {property_name}: array of {property_type}{suffix}")
                        else:
                            _print_output(f"{spaces}- {property_name}: array{suffix}")
                    else:
                        _print_output(f"{spaces}- {property_name}: array{suffix}")
                    if enumeration:
                        nenums = 0
                        maxenums = 15
                        for enum in sorted(enumeration):
                            if (nenums := nenums + 1) >= maxenums:
                                if (remaining := len(enumeration) - nenums) > 0:
                                    _print_output(f"{spaces}  - [{remaining} more ...]")
                                break
                            _print_output(f"{spaces}  - {enum}")
                else:
                    if isinstance(property_type, list):
                        property_type = " or ".join(sorted(property_type))
                    suffix = ""
                    if (enumeration := property.get("enum")) is not None:
                        suffix += f" | enum"
                    if property_required:
                        suffix += f" | required"
                    if property_name in (identifying_properties or []):
                        suffix += f" | identifying"
                    if property.get("uniqueKey"):
                        suffix += f" | unique"
                    if pattern := property.get("pattern"):
                        suffix += f" | pattern: {pattern}"
                    if (format := property.get("format")) and (format != "uuid"):
                        suffix += f" | format: {format}"
                    if isinstance(any_of := property.get("anyOf"), list):
                        if ((any_of == [{"format": "date"}, {"format": "date-time"}]) or
                            (any_of == [{"format": "date-time"}, {"format": "date"}])):  # noqa
                            # Very special case.
                            suffix += f" | format: date or date-time"
                    if link_to := property.get("linkTo"):
                        suffix += f" | reference: {link_to}"
                    if property.get("calculatedProperty"):
                        suffix += f" | calculated"
                    if (default := property.get("default")) is not None:
                        suffix += f" | default:"
                        if isinstance(default, dict):
                            suffix += f" object"
                        elif isinstance(default, list):
                            suffix += f" array"
                        else:
                            suffix += f" {default}"
                    if (minimum := property.get("minimum")) is not None:
                        suffix += f" | min: {minimum}"
                    if (maximum := property.get("maximum")) is not None:
                        suffix += f" | max: {maximum}"
                    if (max_length := property.get("maxLength")) is not None:
                        suffix += f" | max length: {max_length}"
                    if (min_length := property.get("minLength")) is not None:
                        suffix += f" | min length: {min_length}"
                    _print_output(f"{spaces}- {property_name}: {property_type}{suffix}")
                    if enumeration:
                        nenums = 0
                        maxenums = 15
                        for enum in sorted(enumeration):
                            if (nenums := nenums + 1) >= maxenums:
                                if (remaining := len(enumeration) - nenums) > 0:
                                    _print_output(f"{spaces}  - [{remaining} more ...]")
                                break
                            _print_output(f"{spaces}  - {enum}")
            else:
                _print_output(f"{spaces}- {property_name}")


def _print_all_schema_names(portal: Portal,
                            terse: bool = False, all: bool = False,
                            tree: bool = False, summary: bool = False, yaml: bool = False) -> None:
    if not (schemas := _get_schemas(portal)):
        return

    if summary is not True:
        if yaml:
            _print_output(yaml.dump(schemas))
        else:
            _print_output(json.dumps(schemas, indent=4))
        return

    if tree:
        _print_schemas_tree(schemas)
        return

    for schema_name in sorted(schemas.keys()):
        if parent_schema_name := _get_parent_schema_name(schemas[schema_name]):
            if schemas[schema_name].get("isAbstract") is True:
                _print_output(f"{schema_name} | parent: {parent_schema_name} | abstract")
            else:
                _print_output(f"{schema_name} | parent: {parent_schema_name}")
        else:
            if schemas[schema_name].get("isAbstract") is True:
                _print_output(f"{schema_name} | abstract")
            else:
                _print_output(schema_name)
        if not terse:
            _print_schema(schemas[schema_name], terse=terse, all=all)


def _get_parent_schema_name(schema: dict) -> Optional[str]:
    if (isinstance(schema, dict) and
        (parent_schema_name := schema.get("rdfs:subClassOf")) and
        (parent_schema_name := parent_schema_name.replace("/profiles/", "").replace(".json", "")) and
        (parent_schema_name != "Item")):  # noqa
        return parent_schema_name
    return None


def _print_schemas_tree(schemas: dict) -> None:
    def children_of(name: str) -> List[str]:
        nonlocal schemas
        children = []
        if not (name is None or isinstance(name, str)):
            return children
        if name and name.lower() == "schemas":
            name = None
        for schema_name in (schemas if isinstance(schemas, dict) else {}):
            if _get_parent_schema_name(schemas[schema_name]) == name:
                children.append(schema_name)
        return sorted(children)
    def name_of(name: str) -> str:  # noqa
        nonlocal schemas
        if not (name is None or isinstance(name, str)):
            return name
        if (schema := schemas.get(name)) and schema.get("isAbstract") is True:
            return f"{name} (abstact)"
        return name
    _print_tree(root_name="Schemas", children_of=children_of, name_of=name_of)


def _print_tree(root_name: Optional[str],
                children_of: Callable,
                has_children: Optional[Callable] = None,
                name_of: Optional[Callable] = None) -> None:
    """
    Recursively prints as a tree structure the given root name and any of its
    children (again, recursively) as specified by the given children_of callable;
    the has_children may be specified, for efficiency, though if not specified
    it will use the children_of function to determine this; the name_of callable
    may be specified to modify the name before printing.
    """
    first = "└─ "
    space = "    "
    branch = "│   "
    tee = "├── "
    last = "└── "

    if not callable(children_of):
        return
    if not callable(has_children):
        has_children = lambda name: children_of(name) is not None  # noqa

    # This function adapted from stackoverflow.
    # Ref: https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python
    def tree_generator(name: str, prefix: str = ""):
        contents = children_of(name)
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            yield prefix + pointer + (name_of(path) if callable(name_of) else path)
            if has_children(path):
                extension = branch if pointer == tee else space
                yield from tree_generator(path, prefix=prefix+extension)
    _print_output(first + ((name_of(root_name) if callable(name_of) else root_name) or "root"))
    for line in tree_generator(root_name, prefix="   "):
        _print_output(line)


def _read_json_from_file(file: str) -> Optional[dict]:
    if not os.path.exists(file):
        _print(f"Cannot find file: {file}")
        _exit(1)
    try:
        with io.open(file, "r") as f:
            try:
                return json.load(f)
            except Exception:
                _print(f"Cannot parse JSON in file: {file}")
                _exit(1)
    except Exception as e:
        _print(e)
        _print(f"Cannot open file: {file}")
        _exit(1)


def _print(*args, **kwargs):
    with uncaptured_output():
        PRINT(*args, **kwargs)
    sys.stdout.flush()


def _print_output(value: str):
    global _output_file
    if _output_file:
        _output_file.write(value)
        _output_file.write("\n")
    else:
        with uncaptured_output():
            PRINT(value)
        sys.stdout.flush()


def _exit(message: Optional[Union[str, int]] = None, status: Optional[int] = None) -> None:
    global _output_file
    if isinstance(message, str):
        _print(f"ERROR: {message}")
    elif isinstance(message, int) and not isinstance(status, int):
        status = message
    if _output_file:
        _output_file.close()
    sys.exit(status if isinstance(status, int) else (0 if status is None else 1))


if __name__ == "__main__":
    main()



from __future__ import annotations
from functools import lru_cache
from typing import Any, Optional
from hms_utils.chars import chars
from hms_utils.dictionary_parented import JSON
from hms_utils.dictionary_print_utils import print_dictionary_list, print_dictionary_tree
from hms_utils.config.config import Config
from hms_utils.config.config_with_aws_macros import ConfigWithAwsMacros
from hms_utils.terminal_utils import terminal_color


class ConfigOutput:

    @staticmethod
    def print_tree(config: Config,
                   data: Optional[JSON] = None,
                   show: Optional[bool] = False,
                   nocolor: bool = False,
                   indent: Optional[int] = None,
                   string: bool = False,
                   root: Optional[str, bool] = False,
                   debug: bool = False) -> Optional[str]:
        output_to_string = ""
        def print_to_string(value: str) -> None:  # noqa
            nonlocal output_to_string
            output_to_string += value + "\n"
        def value_modifier(path: str, value: Any) -> Optional[str]:  # noqa
            nonlocal config, show
            if show is not None:
                value = ConfigOutput._lookup(config, path, show=None)
                if isinstance(config, ConfigWithAwsMacros) and config._contains_aws_secret_values(value):
                    aws_account_number, aws_secrets_name, aws_secret_name = config._secrets_plaintext_info(value)
                    if aws_account_number:
                        value += (f" {chars.dot} {aws_account_number}:{aws_secrets_name}/{aws_secret_name}")
                return ConfigOutput._display_value(config, value, show=show, nocolor=nocolor)
            elif show is True:
                return ConfigOutput._lookup(config, path, show=show)
        def tree_arrow_indicator(path: str, value: Any) -> Optional[str]:  # noqa
            nonlocal config, show
            if show is not None:
                if config._contains_secret_values(ConfigOutput._lookup(config, path, show=None)):
                    return terminal_color(chars.rarrow, "red", bold=True, nocolor=nocolor)
        if not isinstance(config, Config):
            return
        if not isinstance(data, JSON):
            data = config.data(show=None)
        if root is True:
            root = "root"
        elif not (isinstance(root, str) and root):
            root = None
        if root:
            if debug is True:
                print(f"{chars.rarrow} {root} {chars.dot} {id(data)}")
            else:
                print(f"{chars.rarrow} {root}")
            indent = 2
        print_dictionary_tree(data,
                              value_modifier=value_modifier,
                              arrow_indicator=tree_arrow_indicator,
                              indent=indent,
                              printf=print_to_string if string is True else None,
                              debug=debug)
        if string is True:
            if output_to_string.endswith("\n"):
                output_to_string = output_to_string[:-1]
            return output_to_string

    @staticmethod
    def print_list(config: Config, show: Optional[bool] = False, nocolor: bool = False) -> None:
        def value_modifier(path: str, value: Any) -> Optional[str]:
            nonlocal config, nocolor, show
            if show is not None:
                value = ConfigOutput._lookup(config, path, show=None)
                aws_account_number = aws_secrets_name = aws_secret_name = None
                if isinstance(config, ConfigWithAwsMacros) and config._contains_aws_secret_values(value):
                    aws_account_number, aws_secrets_name, aws_secret_name = \
                        config._secrets_plaintext_info(value)
                value = ConfigOutput._display_value(config, value, show=show, nocolor=nocolor)
                if aws_account_number:
                    value += (f" {chars.dot} {aws_account_number}:{aws_secrets_name}/{aws_secret_name}")
            # elif show is True:
            else:
                value = ConfigOutput._lookup(config, path, show=None)
            return value
        print_dictionary_list(config.data(show=None), value_modifier=value_modifier)

    @staticmethod
    def _display_value(config: Config, value: Any, show: bool, nocolor: bool = False) -> Optional[str]:
        def display_secret_value(value: Any) -> str:
            nonlocal nocolor
            return terminal_color(str(value), "red", bold=True, nocolor=nocolor)
        if show is True:
            return config._secrets_plaintext(value, plaintext_value=display_secret_value)
        elif show is False:
            return config._secrets_obfuscated(value, obfuscated_value=display_secret_value)
        else:
            return value

    @lru_cache
    def _lookup(config: Config, path: str, show: Optional[bool]) -> Any:
        return config.lookup(path, show=show)



from __future__ import annotations
from prettytable import PrettyTable
from typing import Callable, Optional
from hms_utils.chars import chars
from hms_utils.dictionary_parented import JSON


def print_dictionary_tree(data: dict,
                          indent: Optional[int] = None,
                          paths: bool = False,
                          path_separator: str = "/",
                          root_indicator: Optional[Callable] = None,
                          parent_annotator: Optional[Callable] = None,
                          value_annotator: Optional[Callable] = None,
                          key_modifier: Optional[Callable] = None,
                          value_modifier: Optional[Callable] = None,
                          arrow_indicator: Optional[Callable] = None,
                          printf: Optional[Callable] = None,
                          debug: bool = False) -> None:
    """
    Pretty prints the given dictionary. ONLY handles dictionaries
    containing primitive values or other dictionaries recursively.
    """
    if not callable(root_indicator):
        root_indicator = None
    if not callable(parent_annotator):
        parent_annotator = None
    if not callable(value_annotator):
        value_annotator = None
    if not callable(key_modifier):
        key_modifier = None
    if not callable(value_modifier):
        value_modifier = None
    if not callable(arrow_indicator):
        arrow_indicator = None
    if not callable(printf):
        printf = print
    if not isinstance(indent, int):
        indent = 0
    output = (lambda value: printf(f"{' ' * indent}{value}")) if indent > 0 else printf
    def traverse(data: dict, indent: str = "", first: bool = False, last: bool = True, path: str = ""):  # noqa
        if not path:
            if isinstance(data, JSON):
                path = data.path
        nonlocal output, paths, key_modifier, value_annotator, value_modifier
        space = "    " if not first else "  "
        for index, key in enumerate(keys := list(data.keys())):
            last = (index == len(keys) - 1)
            corner = "▷" if first else ("└──" if last else "├──")
            if isinstance(data, JSON):
                key_path = data.context_path(path_separator=path_separator, path_suffix=key, path_rooted=True)
            else:
                key_path = f"{path}{path_separator}{key}" if path else key
            # key_path = f"{path}{path_separator}{key}" if path else key
            if isinstance(value := data[key], dict):
                if (debug is True) and isinstance(value, JSON):
                    key += f" {chars.dot} id: {id(value)}"
                    if parent := value.parent:
                        key += f" {chars.dot} parent: {id(parent)}"
                if parent_annotator and (parent_annotation := parent_annotator(value)):
                    key = key + parent_annotation
                output(indent + corner + " " + key)
                inner_indent = indent + (space if last else f"{' ' if first else '│'}{space[1:]}")
                traverse(value, indent=inner_indent, last=last, path=key_path)
            elif isinstance(value, list):
                output(indent + corner + " " + key)
                inner_indent = (indent if first else indent + (2 * " ")) + (2 * " ")
                for element_index, element_value in enumerate(value):
                    if isinstance(element_value, dict):
                        element_output_value = "{}"
                        if parent_annotator and (parent_annotation := parent_annotator(element_value)):
                            element_output_value += parent_annotation
                        output(f"{inner_indent}└── [{element_index}]: {element_output_value}")
                        traverse(element_value, indent=inner_indent + (9 * " "))
                    else:
                        output(f"{inner_indent}└── [{element_index}]: {element_value}")
            else:
                if paths:
                    key = key_path
                key_modification = key_modifier(key_path, key) if key_modifier else key
                value_modification = value_modifier(key_path, value) if value_modifier else value
                value_annotation = value_annotator(data, key, value) if value_annotator else ""
                arrow_indication = arrow_indicator(key_path, value) if arrow_indicator else ""
                if key_modification:
                    key = key_modification
                if value_modification:
                    value = value_modification
                if arrow_indication:
                    corner = corner[:-1] + arrow_indication
                if (debug is True) and isinstance(data, JSON):
                    if not value_annotation:
                        value_annotation = ""
                    value_annotation += f" {chars.dot} parent: {id(data)}"
                output(f"{indent}{corner} {key}: {value}{f'{value_annotation}' if value_annotation else ''}")
    if root_indicator:
        if isinstance(root_indication := root_indicator(), str) and root_indication:
            output(root_indication)
    traverse(data, first=True)


def print_dictionary_list(data: dict,
                          path_separator: str = "/",
                          prefix: Optional[str] = None,
                          key_modifier: Optional[Callable] = None,
                          value_modifier: Optional[Callable] = None,
                          value_annotator: Optional[Callable] = None) -> None:
    if not callable(key_modifier):
        key_modifier = None
    if not callable(value_annotator):
        value_annotator = None
    if not callable(value_modifier):
        value_modifier = None
    if not isinstance(prefix, str):
        prefix = ""
    def traverse(data: dict, path: str = "") -> None:  # noqa
        nonlocal path_separator, key_modifier, value_annotator, value_modifier
        for key in data:
            key_path = f"{path}{path_separator}{key}" if path else key
            if isinstance(item := data[key], dict):
                traverse(item, path=key_path)
            else:
                value = str(item)
                key_modification = key_modifier(key_path) if key_modifier else key_path
                value_modification = value_modifier(key_path, value) if value_modifier else value
                value_annotation = value_annotator(key_path) if value_annotator else ""
                if key_modification:
                    key = key_modification
                if value_modification:
                    value = value_modification
                print(f"{prefix}{key}: {value}{f' {value_annotation}' if value_annotation else ''}")
    traverse(data)


def print_dictionary_as_table(header_name: str, header_value: str,
                              dictionary: dict, display_value: Callable, sort: bool = True) -> None:
    table = PrettyTable()
    table.field_names = [header_name, header_value]
    table.align[header_name] = "l"
    table.align[header_value] = "l"
    if not callable(display_value):
        display_value = lambda _, value: value  # noqa
    for key_name, key_value in sorted(dictionary.items(), key=lambda item: item[0]) if sort else dictionary.items():
        table.add_row([key_name, display_value(key_name, key_value)])
    print(table)


def print_grouped_items(grouped_items: dict, noitems: bool = False,
                        title: Optional[str] = None,
                        map_grouped_item: Optional[Callable] = None,
                        remove_prefix_grouping_value: bool = False,
                        indent: Optional[int] = None) -> None:
    if not (isinstance(grouped_items, dict) and grouped_items):
        return
    if not (isinstance(indent, int) and (indent > 0)):
        indent = 0
    spaces = (" " * indent) if indent > 0 else ""
    if isinstance(title, str) and title:
        print(f"{spaces}{chars.rarrow} {title}")
        indent += 2
        spaces = " " * indent
    if not (group := grouped_items.get("group")):
        return
    if not (group_count := grouped_items.get("group_count")):
        return
    if not (group_items := grouped_items.get("group_items")):
        return
    if not callable(map_grouped_item):
        map_grouped_item = None
    message = f"{spaces}{chars.diamond} GROUP: {group} ({group_count})"
    print(message)
    for group_item_key in group_items:
        grouped_items = group_items[group_item_key]
        if (remove_prefix_grouping_value is True) and isinstance(group_item_key, str):
            if (colon_index := group_item_key.find(":")) > 0:
                group_item_key = group_item_key[colon_index + 1:]
        message = (f"{spaces}  {chars.rarrow if indent == 0 else chars.rarrow_hollow}"
                   f" {group_item_key if group_item_key is not None else chars.null}")
        if isinstance(grouped_items, dict):
            if isinstance(grouped_items_count := grouped_items.get("item_count"), int):
                message += f" ({grouped_items_count})"
            print(message)
            print_grouped_items(grouped_items, noitems=noitems, map_grouped_item=map_grouped_item,
                                remove_prefix_grouping_value=remove_prefix_grouping_value, indent=indent+4)
        elif isinstance(grouped_items, list):
            print(f"{message} ({len(grouped_items)})")
            if noitems is not True:
                for grouped_item in grouped_items:
                    if map_grouped_item:
                        grouped_item = map_grouped_item(grouped_item)
                    print(f"{spaces}    {chars.dot} {grouped_item}")
        elif isinstance(grouped_items, int):
            print(f"{message} ({grouped_items})")



# On 2024-10-31 it was noticed that smaht-submitr had not been setting consortia for items submitted to smaht-portal.
# This bug has been fixed (C4-1186) and this script (C4-1187) is to correct the existing data.
# Before this was run a backup was made of the smaht-portal (RDS) database: rds-smaht-production-snapshot-20241031

from typing import List
from dcicutils.ff_utils import search_metadata
from hms_utils.portal.portal_utils import Portal
from hms_utils.argv import ARGV
from hms_utils.chars import chars
from hms_utils.threading_utils import run_concurrently

nerrors = 0
total = 0


def main():

    argv = ARGV({
        ARGV.REQUIRED(str, "smaht-data"): ["--env"],
        ARGV.OPTIONAL(bool): ["--dryrun"],
        # Running into Internal Server Error on some Library types - this "fixes" it
        ARGV.OPTIONAL(bool, True): ["--skip-links"],
        ARGV.OPTIONAL(int, 0): ["--limit"]
    })

    portal = Portal(argv.env)

    query = "/search/?type=Item&consortia.display_title=No+value"
    ignored_types = ["AccessKey", "TrackingItem", "Consortium", "SubmissionCenter"]
    results = search_metadata(query, key=portal.key, is_generator=True)
    consortia = ["smaht"]

    print(f"SETTING CONSORTIA FOR PORTAL ITEMS WITHOUT IT: {portal.server}")
    if argv.dryrun:
        print(f"{chars.rarrow}{chars.rarrow}{chars.rarrow} DRY RUN {chars.larrow}{chars.larrow}{chars.larrow}")
    concurrency = 50
    functions = []
    nitems = 0
    for item in results:
        if isinstance(item, dict) and (item_uuid := item.get("uuid")):
            if (item_type := portal.get_schema_type(item)) not in ignored_types:
                if item.get("consortia", None) is None:
                    if (argv.limit > 0) and (nitems >= argv.limit):
                        print(f"REACHED LIMIT: {argv.limit}")
                        break
                    nitems += 1
                    print(f"{item_uuid}: {item_type} {chars.rarrow_hollow} SETTING CONSORTIA: {consortia}", flush=True)
                    if not argv.dryrun:
                        functions.append(lambda portal=portal, item=item, consortia=consortia:
                                         _set_consortia(portal, item, consortia, skip_links=argv.skip_links))
                        if len(functions) >= concurrency:
                            run_concurrently(functions, nthreads=concurrency)
                            functions = []
                else:
                    print(f"{item_uuid}: {item_type} {chars.rarrow_hollow} CONSORTIA ALREADY SET")

    if functions:
        run_concurrently(functions, nthreads=concurrency)

    global total, nerrors
    if argv.dryrun:
        print(f"DONE REVIEWING PORTAL ITEMS WITHOUT CONSORTIA:"
              f" {portal.server} {chars.dot} total: {nitems} {chars.dot} errors: {nerrors}")
        print(f"{chars.rarrow}{chars.rarrow}{chars.rarrow} DRY RUN {chars.larrow}{chars.larrow}{chars.larrow}")
    else:
        print(f"DONE SETTING CONSORTIA FOR PORTAL ITEMS WITHOUT IT:"
              f" {portal.server} {chars.dot} total: {total} {chars.dot} errors: {nerrors}")


def _set_consortia(portal: Portal, item: dict, consortia: List[str], skip_links: bool = False) -> None:
    global total, nerrors
    try:
        item_uuid = item.get("uuid")
        item_type = portal.get_schema_type(item)
        print(f"{item_uuid}: {item_type} {chars.rarrow_hollow} SETTING CONSORTIA: {consortia}", flush=True)
        if skip_links:
            item_uuid += "?skip_links=true"
        portal.patch_metadata(item_uuid, {"consortia": consortia})
        total += 1
    except Exception as e:
        nerrors += 1
        print(e, flush=True)


if __name__ == "__main__":
    main()
    pass



import calendar
from datetime import date, datetime
from dateutil.relativedelta import relativedelta
import pyramid
from typing import Any, Callable, List, Optional, Tuple, Union
from urllib.parse import urlencode
from dcicutils.datetime_utils import parse_datetime_string
from encoded.root import SMAHTRoot

from snovault.search.search import search as snovault_search
from snovault.search.search_utils import make_search_subreq as snovault_make_search_subreq

AGGREGATION_FIELD_RELEASE_DATE = "file_status_tracking.released"
AGGREGATION_FIELD_CELL_LINE = "file_sets.libraries.analytes.samples.sample_sources.cell_line.code"
AGGREGATION_FIELD_DONOR = "donors.display_title"
AGGREGATION_FIELD_FILE_DESCRIPTOR = "release_tracker_description"

AGGREGATIONS = [
    AGGREGATION_FIELD_RELEASE_DATE,
    AGGREGATION_FIELD_CELL_LINE,
    AGGREGATION_FIELD_DONOR,
    AGGREGATION_FIELD_FILE_DESCRIPTOR
]

AGGREGATIONS_BY_CELL_LINE = [
    AGGREGATION_FIELD_RELEASE_DATE,
    AGGREGATION_FIELD_CELL_LINE,
    AGGREGATION_FIELD_FILE_DESCRIPTOR
]

AGGREGATIONS_BY_DONOR = [
    AGGREGATION_FIELD_RELEASE_DATE,
    AGGREGATION_FIELD_DONOR,
    AGGREGATION_FIELD_FILE_DESCRIPTOR
]

AGGREGATION_MAX_BUCKETS = 25
AGGREGATION_NO_VALUE = "No value"

DEFAULT_FILE_TYPES = ["OutputFile"]
DEFAULT_FILE_STATUSES = ["released"]
DEFAULT_FILE_CATEGORIES = ["!Quality Control"]
DEFAULT_RECENT_MONTHS = 3
DEFAULT_INCLUDE_CURRENT_MONTH = True


def recent_files_summary(context: SMAHTRoot, request: pyramid.request.Request) -> dict:

    status = request_args(request, "status", None)
    category = request_args(request, "category", None)
    from_date = request_arg(request, "from", None)
    thru_date = request_arg(request, "thru", None)
    nmonths = request_arg_int(request, "months", DEFAULT_RECENT_MONTHS)
    include_current_month = request_arg_bool(request, "include_current_month", False)

    if isinstance(aggregations := request.params.getall("aggregation"), list) and aggregations:
        results = recent_files_summary_by(aggregations)
    else:
        # Default case: Do these two concurrently (TODO) ...
        results_by_cell_line = recent_files_summary_by(
            request,
            aggregations=AGGREGATIONS_BY_CELL_LINE,
            status=status, category=category,
            from_date=from_date, thru_date=thru_date, nmonths=nmonths,
            include_current_month=include_current_month)
        results_by_donor = recent_files_summary_by(
            request,
            aggregations=AGGREGATIONS_BY_DONOR,
            status=status, category=category,
            from_date=from_date, thru_date=thru_date, nmonths=nmonths,
            include_current_month=include_current_month)
        results = combine_search_aggregation_results(results_by_cell_line, results_by_donor)
    return results


def recent_files_summary_by(request: pyramid.request.Request,
                            aggregations: List[str],
                            status: Optional[str],
                            category: Optional[str],
                            from_date: Optional[str],
                            thru_date: Optional[str],
                            nmonths: Optional[str],
                            include_current_month: bool = False):
    query_string, aggregation_definitions = create_query_for_files(type=DEFAULT_FILE_TYPES,
                                                                   status=DEFAULT_FILE_STATUSES,
                                                                   category=DEFAULT_FILE_CATEGORIES,
                                                                   nmonths=DEFAULT_RECENT_MONTHS,
                                                                   include_current_month=DEFAULT_INCLUDE_CURRENT_MONTH,
                                                                   date_property_name=AGGREGATION_FIELD_RELEASE_DATE,
                                                                   aggregation=aggregations,
                                                                   max_buckets=AGGREGATION_MAX_BUCKETS,
                                                                   missing_value=AGGREGATION_NO_VALUE)

    query = f"/search/?{query_string}"
    request = snovault_make_search_subreq(request, path=query, method="GET")
    results = snovault_search(None, request, custom_aggregations=aggregation_definitions)
    results_normalized = normalize_elastic_search_aggregation_results(results.get("aggregations"))
    print(json.dumps(results_normalized, indent=4))
    return results_normalized


def create_query_for_files(
        type: Optional[Union[str, List[str]]] = None,
        status: Optional[Union[str, List[str]]] = None,
        category: Optional[Union[str, List[str]]] = None,
        from_date: Optional[str] = None,
        thru_date: Optional[str] = None,
        nmonths: Optional[str] = None,
        include_current_month: bool = False,
        date_property_name: Optional[str] = None,
        aggregation: Optional[Union[str, List[str], dict]] = None,
        max_buckets: Optional[int] = None,
        missing_value: Optional[str] = None) -> Tuple[Optional[str], Optional[dict]]:

    global AGGREGATION_FIELD_RELEASE_DATE

    types = []
    if type is None:
        type = DEFAULT_FILE_TYPES
    if not isinstance(type, list):
        type = [type]
    for item in type:
        if isinstance(item, str) and (item := item.strip()) and (item not in types):
            types.append(item)

    statuses = []
    if status is None:
        status = DEFAULT_FILE_STATUSES
    if not isinstance(status, list):
        status = [status]
    for item in status:
        if isinstance(item, str) and (item := item.strip()) and (item not in statuses):
            statuses.append(item)

    categories = []
    if category is None:
        category = DEFAULT_FILE_CATEGORIES
    if not isinstance(category, list):
        category = [category]
    for item in category:
        if isinstance(item, str) and (item := item.strip()) and (item not in categories):
            categories.append(item)

    from_date, thru_date = parse_date_related_arguments(from_date, thru_date, nmonths=nmonths,
                                                        include_current_month=include_current_month, strings=True)
    if from_date or thru_date:
        if not (isinstance(date_property_name, str) and (date_property_name := date_property_name.strip())):
            date_property_name = AGGREGATION_FIELD_RELEASE_DATE

    if not isinstance(aggregation, dict):
        aggregations = []
        if not isinstance(aggregation, list):
            aggregation = [aggregation]
        for item in aggregation:
            if isinstance(item, str) and (item := item.strip()) and (item not in aggregations):
                aggregations.append(item)
        def create_field_aggregation(field: str) -> Optional[dict]:  # noqa
            if field == AGGREGATION_FIELD_RELEASE_DATE:
                return {
                    "date_histogram": {
                        "field": f"embedded.{field}",
                        "calendar_interval": "month",
                        "format": "yyyy-MM",
                        "missing": "1970-01",
                        "order": {"_key": "desc"}
                    }
                }
        aggregation_definitions = create_elasticsearch_aggregation_definitions(
            aggregations,
            max_buckets=max_buckets,
            missing_value=missing_value,
            create_field_aggregation=create_field_aggregation)
    else:
        aggregation_definitions = aggregation

    query_parameters = {
        "type": types if types else None,
        "status": status if status else None,
        "data_category": categories if categories else None,
        f"{date_property_name}.from": from_date if from_date else None,
        f"{date_property_name}.to": thru_date if from_date else None,
        "from": 0,
        "limit": 0,

    }
    query_parameters = {key: value for key, value in query_parameters.items() if value is not None}

    query_string = urlencode(query_parameters, True)
    # Hackishness to change "=!" to "!=" in search_param_lists value for e.g.: "data_category": ["!Quality Control"]
    query_string = query_string.replace("=%21", "%21=")

    return query_string, aggregation_definitions


def create_elasticsearch_aggregation_definitions(fields: List[str],
                                                 max_buckets: Optional[int] = None,
                                                 missing_value: Optional[str] = None,
                                                 create_field_aggregation: Optional[Callable] = None) -> dict:

    global AGGREGATION_MAX_BUCKETS, AGGREGATION_NO_VALUE

    if not (isinstance(fields, list) and fields and isinstance(field := fields[0], str) and field):
        return {}
    if not isinstance(missing_value, str):
        missing_value = AGGREGATION_NO_VALUE
    if not (isinstance(max_buckets, int) and (max_buckets > 0)):
        max_buckets = AGGREGATION_MAX_BUCKETS

    if not (callable(create_field_aggregation) and
            isinstance(field_aggregation := create_field_aggregation(field), dict)):
        field_aggregation = {
            "terms": {
                "field": f"embedded.{field}.raw",
                "missing": missing_value,
                "size": max_buckets
            }
        }

    aggregation = {field: field_aggregation}
    aggregation[field]["meta"] = {"field_name": field}

    if nested_aggregation := create_elasticsearch_aggregation_definitions(
            fields[1:], max_buckets=max_buckets,
            missing_value=missing_value,
            create_field_aggregation=create_field_aggregation):
        aggregation[field]["aggs"] = nested_aggregation

    return aggregation


def normalize_elastic_search_aggregation_results(data: dict) -> dict:
    def get_first_field_with_buckets_list_property(data: dict) -> Optional[dict]:  # noqa 
        if isinstance(data, dict):
            for key in data:
                if isinstance(data[key], dict) and isinstance(data[key].get("buckets"), list):
                    return data[key]
        return None
    def process_field(field: dict) -> None:  # noqa
        if not isinstance(field, dict):
            return
        if not isinstance(buckets := field.get("buckets"), list):
            return
        group_name = field.get("meta", {}).get("field_name")
        group_items = {}
        item_count = 0
        for bucket in buckets:
            if not (key := bucket.get("key_as_string")):
                if (key := bucket.get("key")) in ["No value", "null", "None", None]:
                    key = None
            doc_count = bucket["doc_count"]
            item_count += doc_count
            if nested_field := get_first_field_with_buckets_list_property(bucket):
                group_items[key] = process_field(nested_field)
            else:
                group_items[key] = doc_count
        return {
            "group": group_name,
            "item_count": item_count,
            "group_count": len(group_items),
            "group_items": group_items,
        }
    if not isinstance(data, dict):
        return {}
    return process_field(get_first_field_with_buckets_list_property(data))


def combine_search_aggregation_results(*args) -> dict:
    pass


def request_arg(request: pyramid.request.Request, name: str, fallback: Optional[str] = None) -> Optional[str]:
    return value if (value := request.params.get(name, None)) is not None else fallback


def request_arg_int(request: pyramid.request.Request, name: str, fallback: Optional[int] = 0) -> Optional[Any]:
    if (value := request_arg(request, name)) is not None:
        try:
            return int(value)
        except Exception:
            pass
    return fallback


def request_arg_bool(request: pyramid.request.Request, name: str, fallback: Optional[bool] = False) -> Optional[bool]:
    return fallback if (value := request_arg(request, name)) is None else (value == "true")


def request_args(request: pyramid.request.Request, name: str, fallback: Optional[Any] = None) -> Optional[Any]:
    return value if (value := request.params.getall(name)) else fallback


def parse_date_related_arguments(
        from_date: Optional[Union[str, datetime, date]],
        thru_date: Optional[Union[str, datetime, date]],
        nmonths: Optional[Union[str, int]] = None,
        include_current_month: bool = True,
        strings: bool = False) -> Tuple[Optional[Union[str, datetime]], Optional[Union[str, datetime]]]:

    """
    Returns from/thru dates based on the given from/thru date arguments and optional nmonths argument.
    Given dates may be datetime or date objects or strings. Returned dates are datetime objects, or
    if the the given strings arguments is True, then strings (formatted as YYYY-MM-DD).

    If both of the given from/thru dates are specified/valid then those are returned
    and the given nmonths argument is not used.

    If only the given from date is specified then a None thru date is returned, UNLESS the given nmonths
    argument represents a positive integer, in which case the returned thru date will be nmonths months
    subsequent to the given from date; or if the given nmonths represents zero, in which case the
    returned thru date will be the last date of the month of the given from date.

    If only the given thru date is specified then a None from date is returned, UNLESS the given nmonths
    argument represents a negative integer, in which case the returned from date will be nmonths monthss
    previous to the given thru date; or if the given nmonths represents zero, in which case
    the returned from date will be the first date of the month of the given thru date.

    If neither the given from/thru dates are specified then None is returns for both, UNLESS the given
    nmonths arguments represents a non-zero integer, in which case the returned from/thru dates will represent
    the past (absolute value) nmonths months starting with the month previous to the month of "today"; however
    if the include_current_month is True it is rather the past nmonths starting with the month of "today".
    """
    from_date = _parse_datetime_string(from_date, notz=True)
    thru_date = _parse_datetime_string(thru_date, last_day_of_month_if_no_day=True, notz=True)
    if not isinstance(nmonths, int):
        if isinstance(nmonths, str) and (nmonths := nmonths.strip()):
            try:
                nmonths = int(nmonths)
            except Exception:
                nmonths = 0
        else:
            nmonths = 0
    if from_date:
        if (not thru_date) and isinstance(nmonths, int):
            if nmonths > 0:
                thru_date = add_months(from_date, nmonths)
            elif nmonths == 0:
                thru_date = get_last_date_of_month(from_date)
    elif thru_date:
        if isinstance(nmonths, int):
            if nmonths < 0:
                from_date = add_months(thru_date, nmonths)
            elif nmonths == 0:
                from_date = get_first_date_of_month(thru_date)
    elif isinstance(nmonths, int) and ((nmonths := abs(nmonths)) != 0):
        # If no (valid) from/thru dates given, but the absolute value of nmonths is a non-zero integer, then returns
        # from/thru dates for the last nmonths month ending with the last day of month previous to the current month.
        # thru_date = add_months(get_last_date_of_month(), -1)
        thru_date = get_last_date_of_month()
        if include_current_month is not True:
            thru_date = add_months(thru_date, -1)
        from_date = add_months(thru_date, -nmonths)
    if strings is True:
        return (from_date.strftime(f"%Y-%m-%d") if from_date else None,
                thru_date.strftime(f"%Y-%m-%d") if thru_date else None)
    return from_date, thru_date


def get_first_date_of_month(day: Optional[Union[datetime, date, str]] = None) -> datetime:
    """
    Returns a datetime object representing the first day of the month of the given date;
    this given date may be a datetime or date object, or string representing a date or
    datetime; if the given argument is unspecified or incorrect then assumes "today".
    """
    if not (day := _parse_datetime_string(day, notz=True)):
        day = datetime.today().replace(tzinfo=None)
    return day.replace(day=1)


def get_last_date_of_month(day: Optional[Union[datetime, date, str]] = None) -> datetime:
    """
    Returns a datetime object representing the last day of the month of the given date;
    this given date may be a datetime or date object, or string representing a date or
    datetime; if the given argument is unspecified or incorrect then assumes "today".
    """
    if not (day := _parse_datetime_string(day)):
        day = datetime.today().replace(tzinfo=None)
    return datetime(day.year, day.month, calendar.monthrange(day.year, day.month)[1])


def add_months(day: Optional[Union[datetime, date, str]] = None, nmonths: int = 0) -> datetime:
    """
    Returns a datetime object representing the given date with the given nmonths number of months
    added (or substracted if negative) to (or from) that given date.; this given date may be a
    datetime or date object, or string representing a date or datetime; if the given argument
    is unspecified or incorrect then assumes "today".
    """
    if not (day := _parse_datetime_string(day, notz=True)):
        day = datetime.today().replace(tzinfo=None)
    if isinstance(nmonths, int) and (nmonths != 0):
        return day + relativedelta(months=nmonths)
    return day


def _parse_datetime_string(value: Union[str, datetime, date],
                           last_day_of_month_if_no_day: bool = False,
                           notz: bool = False) -> Optional[datetime]:
    """
    Wrapper around dcicutils.datetime_utils.parse_datetime_string to handle a few special cases for convenience.
    """
    last_day_of_month = False
    if not isinstance(value, datetime):
        if isinstance(value, date):
            value = datetime.combine(value, datetime.min.time())
        elif isinstance(value, str):
            if (len(value) == 8) and value.isdigit():
                # Special case to accept for example "20241206" to mean "2024-12-06".
                value = f"{value[0:4]}-{value[4:6]}-{value[6:8]}"
            elif (len(value) == 7) and (value[4] == "-") and value[0:4].isdigit() and value[5:].isdigit():
                # Special case to accept for example "2024-10" to mean "2024-10-01".
                value = f"{value}-01"
                last_day_of_month = last_day_of_month_if_no_day
            elif (len(value) == 7) and (value[2] == "/") and value[0:2].isdigit() and value[3:].isdigit():
                # Special case to accept for example "11/2024" to mean "2024-11-01".
                value = f"{value[3:]}-{value[0:2]}-01"
                last_day_of_month = last_day_of_month_if_no_day
            elif (len(value) == 6) and (value[1] == "/") and value[0:1].isdigit() and value[2:].isdigit():
                # Special case to accept for example "9/2024" to mean "2024-09-01".
                value = f"{value[2:]}-0{value[0:1]}-01"
                last_day_of_month = last_day_of_month_if_no_day
            if not (value := parse_datetime_string(value)):
                return None
        else:
            return None
    value = value.replace(tzinfo=None) if notz is True else value
    if last_day_of_month:
        value = get_last_date_of_month(value)
    return value


import json  # noqa
import os  # noqa
from urllib.parse import unquote  # noqa
from hms_utils.dictionary_utils import get_property, group_items_by_groupings, normalize_elastic_search_aggregation_results  # noqa
from hms_utils.dictionary_print_utils import print_grouped_items  # noqa
from hms_utils.portal.portal_utils import create_pyramid_request_for_testing, portal_custom_search, Portal  # noqa

if False:
    print(f"\nRECENT RELEASED FILES BY DATE & CELL-LINE & DONOR & FILE-DESCRIPTOR ...\n")
    query_string, aggregation_definitions = create_query_for_files(type=DEFAULT_FILE_TYPES,
                                                                   status=DEFAULT_FILE_STATUSES,
                                                                   category=DEFAULT_FILE_CATEGORIES,
                                                                   nmonths=DEFAULT_RECENT_MONTHS,
                                                                   include_current_month=DEFAULT_INCLUDE_CURRENT_MONTH,
                                                                   date_property_name=AGGREGATION_FIELD_RELEASE_DATE,
                                                                   aggregation=AGGREGATIONS,
                                                                   max_buckets=AGGREGATION_MAX_BUCKETS,
                                                                   missing_value=AGGREGATION_NO_VALUE)
    portal = Portal(os.path.expanduser("~/repos/smaht-portal/development.ini"))
    result = portal_custom_search(portal, f"/search/?{query_string}", aggregations=aggregation_definitions)
    result_aggregations = normalize_elastic_search_aggregation_results(result.get("aggregations"))
    print_grouped_items(result_aggregations)

if False:
    print(f"\nRECENT RELEASED FILES BY DATE & CELL-LINE & FILE-DESCRIPTOR ...\n")
    query_string, aggregation_definitions = create_query_for_files(type=DEFAULT_FILE_TYPES,
                                                                   status=DEFAULT_FILE_STATUSES,
                                                                   category=DEFAULT_FILE_CATEGORIES,
                                                                   nmonths=DEFAULT_RECENT_MONTHS,
                                                                   include_current_month=DEFAULT_INCLUDE_CURRENT_MONTH,
                                                                   date_property_name=AGGREGATION_FIELD_RELEASE_DATE,
                                                                   aggregation=AGGREGATIONS_BY_CELL_LINE,
                                                                   max_buckets=AGGREGATION_MAX_BUCKETS,
                                                                   missing_value=AGGREGATION_NO_VALUE)
    portal = Portal(os.path.expanduser("~/repos/smaht-portal/development.ini"))
    result = portal_custom_search(portal, f"/search/?{query_string}", aggregations=aggregation_definitions)
    result_aggregations = normalize_elastic_search_aggregation_results(result.get("aggregations"))
    print_grouped_items(result_aggregations)

if False:
    print(f"\nRECENT RELEASED FILES BY DATE & DONOR & FILE-DESCRIPTOR ...\n")
    query_string, aggregation_definitions = create_query_for_files(type=DEFAULT_FILE_TYPES,
                                                                   status=DEFAULT_FILE_STATUSES,
                                                                   category=DEFAULT_FILE_CATEGORIES,
                                                                   nmonths=DEFAULT_RECENT_MONTHS,
                                                                   include_current_month=DEFAULT_INCLUDE_CURRENT_MONTH,
                                                                   date_property_name=AGGREGATION_FIELD_RELEASE_DATE,
                                                                   aggregation=AGGREGATIONS_BY_DONOR,
                                                                   max_buckets=AGGREGATION_MAX_BUCKETS,
                                                                   missing_value=AGGREGATION_NO_VALUE)
    portal = Portal(os.path.expanduser("~/repos/smaht-portal/development.ini"))
    result = portal_custom_search(portal, f"/search/?{query_string}", aggregations=aggregation_definitions)
    result_aggregations = normalize_elastic_search_aggregation_results(result.get("aggregations"))
    print_grouped_items(result_aggregations)


def test():
    # Arguments to /recent_files_summary endpoint/API.
    request_args = {
        "abc": ["def", 123],
        "ghi": "jkl"
    }
    # Dummy for local testing.
    portal = Portal(os.path.expanduser("~/repos/smaht-portal/development.ini"))
    request = create_pyramid_request_for_testing(portal.vapp, request_args)
    # Call the /recent_files_summary endpoint/API.
    results = recent_files_summary(None, request)
    # Dump the results.
    print(json.dumps(results, indent=4))


test()



class chars:
    check = "✓"
    xmark = "✗"
    dot = "•"
    dot_hollow = "◦"
    diamond = "❖"
    rarrow = "▶"
    larrow = "◀"
    rarrow_hollow = "▷"
    larrow_hollow = "◁"
    square_hollow = "□"
    null = "∅"
